"""
    SquidAgentAdapter - Access squid services using the squid-py api
"""

import logging
import time
import json

from web3 import Web3

from ocean_keeper.account import Account as SquidAccount

from squid_py.config_provider import ConfigProvider
from squid_py.config import Config as SquidConfig
from squid_py.ocean import Ocean as SquidOcean
from ocean_utils.did import (
    did_to_id_bytes,
    did_to_id,
    DID,
)
from squid_py.ocean.keeper import SquidKeeper

from ocean_utils.agreements.service_agreement import ServiceAgreement
from ocean_utils.agreements.service_types import ServiceTypes
from squid_py.brizo.brizo_provider import BrizoProvider
from ocean_keeper.web3_provider import Web3Provider
from squid_py.ocean.ocean_tokens import OceanTokens

from ocean_utils.ddo.metadata import Metadata
from ocean_keeper.agreements.agreement_manager import AgreementStoreManager
from plecos import is_valid_dict_local, validate_dict_local

from starfish.middleware.starfish_events_manager import StarfishEventsManager


logger = logging.getLogger('starfish.squid_agent_adapter')
# from starfish import logger

# to keep the squid_agent_adapter seperate, we will issue a seperate purchase exception just
# from this class

class SquidAgentAdapterPurchaseError(Exception):
    """ Raised when a purchase event has failed to complete """


class AgreementStoreManagerExtra(AgreementStoreManager):
    def get_agreement_ids_for_did(self, did):
        return self.contract_concise.getAgreementIdsForDID(did)

class SquidAgentAdapter():


    def __init__(self, ocean, options=None):
        """init a standard ocean object"""
        self._ocean = ocean
        self._squid_ocean = None

        if not isinstance(options, dict):
            options = {}

        self._aquarius_url = options.get('aquarius_url', 'http://localhost:5000')
        self._brizo_url = options.get('brizo_url', 'http://localhost:8030')
        self._secret_store_url = options.get('secret_store_url', 'http://localhost:12001')
        self._storage_path = options.get('storage_path', 'squid_py.db')
        self._parity_url = options.get('parity_url', self._ocean.keeper_url)


        # clear out any old connections to a different network
        # this means removing the static web3 connection in squid
        Web3Provider._web3 = None

        # make sure we have a instance of squid ocean created before starting
        self.get_squid_ocean()

        # to get past codacy static method 'register_agent'
        self._keeper = SquidKeeper.get_instance()

    def register_asset(
            self, metadata, account,
            service_descriptors=None, providers=None, use_secret_store=True):
        """

        Register an asset with the agent storage server

        :param dict metadata: metadata to write to the storage server
        :param object account: squid account to register the asset
        :param dict options: options to pass to the squid asset create function
        """
        squid_ocean = self.get_squid_ocean(account)
        return squid_ocean.assets.create(
            metadata,
            account,
            service_descriptors,
            providers,
            use_secret_store
            )

    def validate_metadata(self, metadata):
        """

        Validate the metadata with plesto

        :param dict metadata: metadata to validate
        :return: True if the metadata is valid
        :type: boolean

        """
        if self._ocean:
            if is_valid_dict_local(metadata):
                return True
            else:
                validator = validate_dict_local(metadata)
                print(validator)
        return False

    def read_asset(self, did):
        """

        Read the asset metadata(DDO) using the asset DID

        :param str did: DID of the asset to read.

        :return: DDO of the read asset or None if not found
        :type: dict or None

        """
        squid_ocean = self.get_squid_ocean()
        return squid_ocean.assets.resolve(did)

    def search_assets(self, text, sort=None, offset=100, page=1):
        """
        Search assets from the squid API.
        """
        squid_ocean = self.get_squid_ocean()
        if isinstance(text, str):
            ddo_list = squid_ocean.assets.search(text, sort, offset, page)
        elif isinstance(text, dict):
            ddo_list = squid_ocean.assets.query({'query': text}, sort, offset, page)
        else:
            ddo_list = None
        return ddo_list

    def is_service_agreement_template_registered(self, template_id):
        """
        :return: True if the service level agreement template has already been registered
        """
        return not self.get_service_agreement_template_owner(template_id) is None

    def get_service_agreement_template_owner(self, template_id):
        """
        :return: Owner of the registered service level agreement template, if not registered then return None
        """
        return self._keeper.service_agreement.get_template_owner(template_id)

    def register_service_agreement_template(self, template_id, account):
        """
        Try to register service level agreement template using an account

        :param template_id: template id to use to register
        :param account: account to register for
        :return: The template registered


        # template = ServiceAgreementTemplate.from_json_file(get_sla_template_path())
        template = register_service_agreement_template(
            self._keeper.service_agreement,
            account,
            template,
            self._keeper.network_name
        )
        return template
        """
        return None

    @staticmethod
    def get_asset_purchase_ids(did):
        """
        Return a list of purchase id's that have been issued for an asset did
        """
        result = []
        manager = AgreementStoreManagerExtra.get_instance()
        id_list = manager.get_agreement_ids_for_did(did_to_id(did))
        for value in id_list:
            result.append(Web3.toHex(value))
        return result

    def purchase_asset(self, ddo, account):
        """
        Purchase an asset with the agent storage server
        :param dict ddo: ddo of the asset
        :param object account: squid account unlocked and has sufficient funds to buy this asset

        :return: service_agreement_id of the purchase or None if no purchase could be made
        """
        squid_ocean = self.get_squid_ocean(account)

        service_agreement_id = None
        service_agreement = SquidAgentAdapter.get_service_agreement_from_ddo(ddo)
        if service_agreement:
            service_agreement_id = squid_ocean.assets.order(
                ddo.did,
                service_agreement.sa_definition_id,
                account,
                auto_consume=False
            )
        return service_agreement_id

    def purchase_wait_for_completion(self, did, address, purchase_id, timeout_seconds):
        """

        Wait for a purchase to complete

        """

        event = self._keeper.escrow_access_secretstore_template.subscribe_agreement_created(
            purchase_id,
            timeout_seconds,
            SquidAgentAdapter.log_event(self._keeper.escrow_access_secretstore_template.AGREEMENT_CREATED_EVENT),
            (),
            wait=True
        )
        if not event:
            raise SquidAgentAdapterPurchaseError('no event for EscrowAccessSecretStoreTemplate.AgreementCreated')

        event = self._keeper.lock_reward_condition.subscribe_condition_fulfilled(
            purchase_id,
            timeout_seconds,
            SquidAgentAdapter.log_event(self._keeper.lock_reward_condition.FULFILLED_EVENT),
            (),
            wait=True
        )
        if not event:
            raise SquidAgentAdapterPurchaseError('no event for LockRewardCondition.Fulfilled')

        timeout_time = time.time() + timeout_seconds
        while self.is_access_granted_for_asset(did, address, purchase_id) is not True and timeout_time > time.time():
            time.sleep(1)

        return self.is_access_granted_for_asset(did, address, purchase_id)

    def purchase_operation(self, ddo, account):
        """
        Purchase an invoke operation
        :param dict ddo: ddo of the asset
        :param object account: squid account unlocked and has sufficient funds to buy this asset

        :return: service_agreement_id of the purchase or None if no purchase could be made
        """
        squid_ocean = self.get_squid_ocean()

        service_agreement_id = None
        service_agreement = SquidAgentAdapter.get_service_agreement_from_ddo(ddo)
        agreements = squid_ocean.agreements
        did = ddo.did
        if service_agreement:
            logger.info(f'purchase invoke operation ')
            service_definition_id = service_agreement.sa_definition_id

            service_agreement_id, signature = agreements.prepare(ddo.did, service_definition_id, account)
            asset = agreements._asset_resolver.resolve(did)

            service_agreement = ServiceAgreement.from_ddo(service_definition_id, asset)
            # service_def = asset.find_service_by_id(service_definition_id).as_dictionary()

            # Must approve token transfer for this purchase

            agreements._approve_token_transfer(service_agreement.get_price(), account)
            # subscribe to events related to this agreement_id before sending the request.
            logger.debug(f'Registering service agreement with id: {service_agreement_id}')

            BrizoProvider.get_brizo().initialize_service_agreement(did,
                                                                   service_agreement_id,
                                                                   service_agreement.sa_definition_id,
                                                                   signature,
                                                                   account.address,
                                                                   service_agreement.purchase_endpoint)

        return service_agreement_id


    def consume_asset(self, ddo, account, service_agreement_id):
        """
        Conusmer the asset data, by completing the payment and later returning the data for the asset

        """
        result = None
        squid_ocean = self.get_squid_ocean(account)
        service_agreement = SquidAgentAdapter.get_service_agreement_from_ddo(ddo)
        if service_agreement:
            result = json.loads(squid_ocean.secret_store.decrypt(
                    ddo.asset_id,
                    ddo.metadata['base']['encryptedFiles'],
                    account
            ))

#        squid_ocean.assets.consume(service_agreement_id, ddo.did, service_agreement.sa_definition_id, account, download_path)
        return result

    def is_access_granted_for_asset(self, did, account, agreement_id):
        """
        Return true if we have access to the asset's data using the service_agreement_id and account used
        to purchase this asset
        """
        squid_ocean = self.get_squid_ocean()

        account_address = None
        if isinstance(account, object):
            account_address = account.address
        elif isinstance(account, str):
            account_address = account
        else:
            raise TypeError(f'You need to pass an account object or account address')

        return squid_ocean.agreements.is_access_granted(agreement_id, did, account_address)

    def get_purchase_address(self, agreement_id):
        return self._keeper.escrow_access_secretstore_template.get_agreement_consumer(agreement_id)

    def _as_config_dict(self, options=None):
        """

        Return a set of config values, so that squid can read.

        :param options: optional values to add to the dict to send to squid
        :type options: dict or None


        :return: a dict that is compatiable with the current supported version of squid-py.
        :type: dict

        """
        data = {
            'keeper-contracts': {
                'keeper.url': self._ocean.keeper_url,
                'keeper.path': self._ocean.contracts_path,
                'secret_store.url': self._secret_store_url,
                'parity.url': self._parity_url,
            },
            'resources': {
                'aquarius.url': self._aquarius_url,
                'brizo.url': self._brizo_url,
                'storage.path': self._storage_path,
                'downloads.path': '',
            }
        }
        if options:
            if 'parity_address' in options:
                data['keeper-contracts']['parity.address'] = options['parity_address']
            if 'parity_password' in options:
                data['keeper-contracts']['parity.password'] = options['parity_password']
            if 'parity_keyfile' in options:
                data['keeper-contracts']['parity.keyfile'] = options['parity_keyfile']

        return data

    @staticmethod
    def get_account(address, password, keyfile):
        """
        :return: Squid Account object, based on it's address, password and JSON keyfile
        :type: object or None
        """
        return SquidAccount(address, password, keyfile)

    def request_tokens(self, account, value):
        """
        Request some ocean tokens
        :param object account: squid account to request
        :param int value: amount of tokens to request
        :return: number of tokens requested and added to the account
        :type: number
        """
        squid_ocean = self.get_squid_ocean()
        return squid_ocean.accounts.request_tokens(value, account)

    def get_account_balance(self, account):
        """

        :param object account: squid account to get the balance for.
        :return: ethereum and ocean balance of the account
        :type: tuple(eth,ocn)
        """
        squid_ocean = self.get_squid_ocean()
        return squid_ocean.accounts.balance(account)

    @staticmethod
    def create_account_host(password):
        """

        Create a hosted account on the block chain node

        :param str password: password of the new account
        :return: None or squid account of the new account.
        :type: object or None

        """
        account_address = Web3Provider.get_web3().personal.newAccount(password)
        return account_address

    @staticmethod
    def transfer_ether(from_account, to_address, amount):
        tx_hash = Web3Provider.get_web3().personal.sendTransaction({
            'from': from_account.address,
            'to': to_address,
            'value': amount,
        }, from_account.password)
        return Web3Provider.get_web3().eth.waitForTransactionReceipt(tx_hash)

    def transfer_tokens(self, from_account, to_address, amount):
        tokens = OceanTokens(self._keeper.get_instance())
        return tokens.transfer(to_address, amount, from_account)

    def approve_tokens(self, spender_address, price, from_account):
        return self._keeper.token.token_approve(spender_address, price, from_account)

    def register_ddo(self, did, ddo_text, account):
        """register a ddo object on the block chain for this agent

        The account must be an agent_adapter acount
        """
        # register/update the did->ddo to the block chain
        squid_ocean = self.get_squid_ocean()
        checksum = Web3.toBytes(Web3.sha3(ddo_text.encode()))
        did_bytes = did_to_id_bytes(did)
        receipt = squid_ocean._keeper.did_registry.register(did_bytes, checksum, ddo_text, account)

        # transaction = self._squid_ocean._keeper.did_registry._register_attribute(did_id, checksum, ddo_text, account, [])
        # receipt = self._squid_ocean._keeper.did_registry.get_tx_receipt(transaction)
        return receipt

    def resolve_did(self, did):
        """resolve a DID to a given DDO, return the DDO if found"""

        did_bytes = did_to_id_bytes(did)
        data = self._keeper.did_registry.get_registered_attribute(did_bytes)
        if data:
            return data['value']
        return None

    def start_agreement_events_monitor(self, account, callback=None):
        """ called by the publisher to watch payment request events for the published assets
        The account must be an agent_adapter account
        """
        squid_ocean = self.get_squid_ocean(account)

        events_manager = StarfishEventsManager.get_instance(
            squid_ocean._keeper, squid_ocean._config.storage_path, account)

        events_manager.start_agreement_events_monitor(callback)

    def stop_agreement_events_monitor(self, account=None):
        """ called by the publisher to watch payment request events for the published assets
        The account must be an agent_adapter account
        """
        squid_ocean = self.get_squid_ocean(account)

        events_manager = StarfishEventsManager.get_instance(
            squid_ocean._keeper, squid_ocean._config.storage_path, account)

        events_manager.stop_agreement_events_monitor()


    @property
    def accounts(self):
        squid_ocean = self.get_squid_ocean()
        return squid_ocean.accounts.list()

    @property
    def aquarius_url(self):
        return self._aquarius_url

    @property
    def brizo_url(self):
        return self._brizo_url

    def get_squid_ocean(self, account=None):
        """

        Return an instance of squid for an account

        """

        options = {}
        if not self._squid_ocean:
            config_params = self._as_config_dict(options)
            config = SquidConfig(options_dict=config_params)
            ConfigProvider.set_config(config)
            self._squid_ocean = SquidOcean()
        return self._squid_ocean

    @staticmethod
    def get_service_agreement_from_ddo(ddo):
        """
        return the service agreement definition for this asset
        """
        service_agreement = None
        if ddo:
            service = ddo.get_service(service_type=ServiceTypes.ASSET_ACCESS)
            if ServiceAgreement.SERVICE_DEFINITION_ID not in service.as_dictionary():
                raise KeyError(f'cannot find services definition id {ServiceAgreement.SERVICE_DEFINITION_ID}')
            service_agreement = ServiceAgreement.from_service_dict(service.as_dictionary())
        return service_agreement

    @staticmethod
    def get_default_metadata():
        return Metadata.get_example()

    @staticmethod
    def log_event(event_name):
        def _process_event(event):
            logging.debug(f'Received event {event_name}: {event}')
        return _process_event

    @staticmethod
    def generate_did():
        return DID.did()



"""
    test_memory_asset

"""

import secrets
import logging
import json

from starfish.asset import DataAsset




TEST_DATA = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum eu congue odio, vel congue sapien. Morbi ac purus ornare, volutpat elit a, lacinia odio. Integer tempor tellus eget iaculis lacinia. Curabitur aliquam, dui vel vestibulum rhoncus, enim metus interdum enim, in sagittis massa est vel velit. Nunc venenatis commodo libero, vitae elementum nulla ultricies id. Aliquam erat volutpat. Cras eu pretium lacus, quis facilisis mauris. Duis sem quam, blandit id tempor in, porttitor at neque. Cras ut blandit risus. Maecenas vitae sodales neque, eu ultricies nibh.'


def test_init():
    asset = DataAsset.create('test data asset', TEST_DATA)
    assert(asset)
    assert(isinstance(asset, DataAsset))
    assert(asset.data == TEST_DATA.encode('utf-8'))




"""

    Tests for `squid-py issues`.

"""


from tests.integration.squid_issues.example_config import ExampleConfig
from squid_py import Ocean
from squid_py import ConfigProvider

def test_config_setup_working():
    config = ExampleConfig.get_config()
    ConfigProvider.set_config(config)
    ocn = Ocean()

def test_config_setup():

    config = ExampleConfig.get_config()
    ocn = Ocean(config)



"""
    test_bundle_asset

"""

import secrets
import logging
import json

from starfish.asset import DataAsset
from starfish.asset import BundleAsset

TEST_ASSET_COUNT = 4


def test_init():
    bundle = BundleAsset.create('Bundle Asset')
    assert(bundle)

    bundle = BundleAsset({
        'name': 'Bundle Asset',
        'type': 'bundle'
    })
    assert(bundle)

def test_bundle_asset_add_access_remove(ocean, config):
    bundle = BundleAsset.create('name')
    asset_list = {}

    assert(bundle.is_bundle)
    # add a set of memory assets

    for index in range(0, TEST_ASSET_COUNT):
        test_data = secrets.token_hex(1024)
        name = f'name_{index}'
        asset_list[name] = DataAsset.create(f'Asset_{index}', test_data)
        bundle.add(name, asset_list[name])
        assert(bundle[name])
        assert(bundle.asset_count == index + 1)

    # using the iterator
    for name, asset in bundle:
        assert(name)
        assert(asset)
        assert(asset.data == asset_list[name].data)

    # using the get_asset and __getitem__
    for index in range(0, bundle.asset_count):
        asset = bundle.get_asset(index)
        name = list(asset_list.keys())[index]
        assert(asset.data == asset_list[name].data)
        asset = bundle.get_asset(name)
        assert(asset)
        assert(asset.data == asset_list[name].data)
        asset = bundle[name]
        assert(asset)
        assert(asset.data == asset_list[name].data)

        items = dict(bundle.asset_items)
        asset = items[name]
        assert(asset)
        assert(asset.data == asset_list[name].data)



    # test remove
    for name in bundle.asset_names:
        asset = bundle.asset_remove(name)
        assert(asset)

    assert(bundle.asset_count == 0)



"""
    Authentication Class
    To handle embedded public keys
"""

import json
import re

from ocean_utils.ddo.public_key_base import PublicKeyBase


class Authentication:
    """DDO Authentication"""

    def __init__(self, key_id, authentication_type):
        """ init an authentication based on it's key_id and authentication type"""
        self._public_key = None
        self._public_key_id = None
        if isinstance(key_id, PublicKeyBase):
            self._public_key = key_id
        else:
            self._public_key_id = key_id
        self._type = authentication_type

    def assign_did(self, did):
        """
        Assign a DID to the authentication, if the DID does not end with a `#.*`
        then add an automatic key value
        """
        if self._public_key_id:
            if re.match('^#.*', self._public_key_id):
                self._public_key_id = did + self._public_key_id
        if self._public_key:
            self._public_key.assign_did(did)

    def get_type(self):
        """Get the authentication type."""
        return self._type

    def get_public_key_id(self):
        """Get the authentication key id used to validate this authentication."""
        if self._public_key_id:
            return self._public_key_id
        if self._public_key:
            return self._public_key.get_id()
        return None

    def get_public_key(self):
        """Get the authentication public key."""
        return self._public_key

    def as_text(self, is_pretty=False):
        """Return the authentication as a JSON text."""
        values = {
            'type': self._type
        }
        if self._public_key:
            values['publicKey'] = self._public_key.as_text(is_pretty)
        elif self._public_key_id:
            values['publicKey'] = self._public_key_id

        if is_pretty:
            return json.dumps(values, indent=4, separators=(',', ': '))

        return json.dumps(values)

    def as_dictionary(self):
        """Return the authentication as a dictionary."""
        values = {
            'type': self._type
        }
        if self._public_key:
            values['publicKey'] = self._public_key.as_dictionary()
        elif self._public_key_id:
            values['publicKey'] = self._public_key_id

        return values

    def is_valid(self):
        """Return true if this authentication has valid structure."""
        return self.get_public_key_id() is not None and self._type is not None

    def is_public_key(self):
        """Return true if this authentication has an embedded public key."""
        return self._public_key is not None

    def is_key_id(self, key_id):
        """Return True if the `key_id` is the same as this key_id."""
        if self.get_public_key_id() and self.get_public_key_id() == key_id:
            return True
        return False



"""
    test_asset_base

"""

import secrets
import logging
import json

from starfish.asset import AssetBase
from ocean_utils.did import did_to_id

ASSET_METADATA = {
    'name': 'Asset',
    'type': 'asset',
}

TEST_DID = 'did:dep:' + secrets.token_hex(32)

def test_init(metadata):
    asset = AssetBase(ASSET_METADATA)
    assert(asset)
    assert(isinstance(asset, AssetBase))

def test_metadata():
    asset = AssetBase(ASSET_METADATA)
    assert(asset)
    assert(asset.metadata == ASSET_METADATA)

def test_data():
    asset = AssetBase(ASSET_METADATA, TEST_DID)
    assert(asset)
    assert(asset.metadata == ASSET_METADATA)
    assert(asset.did == TEST_DID)


def test_is_asset_type():
    asset = AssetBase(ASSET_METADATA)
    assert(asset)
    assert(asset.is_asset_type('asset'))
    assert(not asset.is_asset_type('bad asset type'))

def test_asset_id():
    asset = AssetBase(ASSET_METADATA, TEST_DID)
    assert(asset)
    asset_id = did_to_id(TEST_DID)
    assert(asset.asset_id == f'0x{asset_id}')


    path_did = 'did:dep:' + secrets.token_hex(32)
    asset_id = did_to_id(TEST_DID)

    asset = AssetBase(ASSET_METADATA, f'{path_did}/{asset_id}')
    assert(asset)
    assert(asset.asset_id == f'0x{asset_id}')



import json
from eth_utils import remove_0x_prefix

from starfish.asset.asset_base import AssetBase
from starfish.asset.data_asset import DataAsset
from starfish.asset.remote_data_asset import RemoteDataAsset
from starfish.asset.bundle_asset import BundleAsset
from starfish.asset.operation_asset import OperationAsset


def create_asset_from_metadata_text(metadata_text, did=None):
    """
    Create a new asset class based on the metadata. Once created assign
    the metadata to the asset and also the optional did

    :param str metadata_text: metadata text to test and create the correct asset object
    :param str did: optional did to assign to the asset

    :return: Return an asset class based on the metadata type
    """

    metadata = json.loads(metadata_text)
    asset_type = AssetBase.get_asset_type(metadata)
    if asset_type == 'bundle':
        return BundleAsset(metadata, did, metadata_text=metadata_text)
    elif asset_type == 'operation':
        return OperationAsset(metadata, did, metadata_text=metadata_text)
    elif asset_type == 'dataset':
        return DataAsset(metadata, did, metadata_text=metadata_text)
    else:
        raise ValueError(f'Unknown asset type {asset_type}')
    return AssetBase(metadata, did, metadata_text=metadata_text)


def is_asset_hash_valid(asset_id, hash_hex):
    """

    :param str asset_id: asset id to check against
    :param str hash_hex: hex string to check
    :return: true if the hash string is the same as the asset_id
    :type: boolean
    """
    if asset_id or hash:
        return remove_0x_prefix(asset_id).lower() == remove_0x_prefix(hash_hex).lower()
    return False



"""

    Command line base output

"""

class BaseOutput():
    """
    Base output to handle the basic formating to Console, JSON and Test
    """

    def __init__(self):
        """init and clear the object"""
        self._header = {}
        self._items = {}
        self._rows = []

    def clear(self):
        """clear the collection of output items"""
        self._items = {}
        self._rows = []
        self._header = {}

    def show(self, text=None, name=None):
        if text is not None:
            if name is None:
                name = 'value'
            self.show_item(name, text)

    def show_item(self, name, value, format_text=None):
        self._items[name] = value

    def set_header(self, name, title, format_text=None):
        self._header[name] = {
            'name': name,
            'title': title,
            'format': format_text
        }
    def show_header(self):
        pass

    def set_item(self, name, value):
        self._items[name] = value

    def flush(self):
        self._rows.append(self._items)

    @property
    def rows(self):
        return self._rows

    @property
    def items(self):
        return self._items

    @property
    def header(self):
        return self._header



"""
    Test invoke agent

"""

import pathlib
import json
import logging
import time

from starfish import (
    Ocean,
    logger
)
from starfish.agent.invoke_agent import InvokeAgent

logger = logging.getLogger('test.invoke')

# koi is not working with the current version of keeper 0.9.0

def _test_invoke():

    agent = InvokeAgent()
    assert agent

    result = agent.get_operations()
    assert 'hashing_did' == result['hashing']
    assert 'echo_did' == result['echo']

    operation = agent.get_operation('echo_did')
    assert operation

    schema = operation.get_schema()
    assert 1 == len(schema)
    assert schema['firstparam'] == 'string'

    result = operation.invoke(firstparam='ocean')
    assert result['firstparam'] == 'hello ocean'





"""
    test_19_asset_download

    As a developer working with an Ocean marketplace,
    I want to record a purchase.

"""

import secrets
import logging
import json

from starfish.asset import DataAsset


def test_19_asset_download(resources, remote_agent):
    test_data = secrets.token_hex(1024)
    asset = DataAsset.create('TestAsset', test_data)
    listing = remote_agent.register_asset(asset, resources.listing_data)

    url = remote_agent.get_asset_store_url(asset.asset_id)
    remote_agent.upload_asset(asset)

    # now download
    store_asset = remote_agent.download_asset(asset.asset_id, url)

    assert(store_asset)
    assert(store_asset.data == asset.data)
    assert(store_asset.asset_id == asset.asset_id)



"""
    SurferAgentAdapter - Adapter to access the Surfer Services
"""
import io
import requests
from starfish import logger
from starfish.utils.crypto_hash import hash_sha3_256

# default base URI for this version surfer
SURFER_BASE_URI = '/api/v1'

SUPPORTED_SERVICES = {
    'metadata': 'DEP.Meta.v1',
    'storage': 'DEP.Storage.v1',
    'invoke': 'DEP.Invoke.v1',
    'market': 'DEP.Market.v1',
    'trust': 'DEP.Trust.v1',
    'auth': 'DEP.Auth.v1',
}

class SurferAgentInvokeAPIGenerator():

    invoke_sync_uri = '/invoke'
    invoke_async_uri = '/invokeasync'
    jobs_uri = '/jobs'
    result = None
    def generate_url(self, endpoint, name):
        if name == 'invoke_sync':
            result = f'{endpoint}{self.invoke_sync_uri}'
        elif name == 'invoke_async':
            result = f'{endpoint}{self.invoke_async_uri}'
        elif name == 'jobs':
            result = f'{endpoint}{self.jobs_uri}'
        else:
            raise ValueError('invalid generation name')
        return result

class ResponseWrapper():
    """
    Response object returned by different test and production
    systems are not the same, and have different properties to obtain
    data and JSON data.

    This class tries to obtain the correct call to get the data from the
    Response object.

    """
    def __init__(self, response):
        self._response = response

    @property
    def json(self):
        if hasattr(self._response, 'get_json'):
            data = self._response.get_json()
        elif hasattr(self._response, 'json'):
            data = self._response.json()
        else:
            TypeError('cannot get the json property from response object')
        return data

    @property
    def data(self):
        if hasattr(self._response, 'content'):
            data = self._response.content
        elif hasattr(self._response, 'data'):
            data = self._response.data
        else:
            raise TypeError('Cannot find correct response data')
        return data



class SurferAgentAdapter():
    _http_client = requests
    _invoke_api_generator = SurferAgentInvokeAPIGenerator()

    def __init__(self, ocean, did=None, ddo=None, options=None):
        """init a standard ocan connection, with a given DID"""
        self._ocean = ocean
        self._did = did
        self._ddo = ddo
        if not options:
            options = {}

#        self._headers = {'content-type': 'application/json'}
        self._headers = {'content-type': 'text/plain'}
        authorization = options.get('authorization')
        if authorization: # this is an OAuth2 token
            self._headers['Authorization'] = f'token {authorization}'

    def register_asset(self, metadata):
        """
        Register an asset with the agent storage server
        :param metadata: metadata to write to the storage server

        :return: A dict of the following items
        [0] asset_id.
        [1] did of the asset.
        [2] metadata in text format, that was saved.

        :type: dict
        """
        result = None
        saved_asset_id = self.save_metadata(metadata)
        result = {
                'asset_id': saved_asset_id,
                'metadata': metadata,
                'hash': hash_sha3_256(metadata)
            }
        return result

    def save_metadata(self, metadata):
        """save metadata to the agent server, using the asset_id and metadata"""
        url = self.get_endpoint('metadata')
        logger.debug(f'metadata save url {url}')
        self._content_header = 'text/plain'
        response = SurferAgentAdapter._http_client.post(url, data=metadata, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            return data
        else:
            msg = f'metadata asset response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def create_listing(self, asset_id, listing_data):
        endpoint = self.get_endpoint('market')
        url = f'{endpoint}/listings'
        data = {
            'assetid': asset_id,
            'info': listing_data,
        }

        self._content_header = 'application/json'
        response = SurferAgentAdapter._http_client.post(url, json=data, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            logger.debug('listing response returned: ' + str(data))
            return data
        else:
            msg = f'listing response failed: {response.status_code} {response.text}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def download_asset(self, url):
        self._content_header = 'application/octet-stream'
#        self._content_header = 'text/plain'
        response = SurferAgentAdapter._http_client.get(url, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).data
            return data
        else:
            msg = f'GET assets response failed: {response.status_code} {response}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def get_asset_store_url(self, asset_id):
        endpoint = self.get_endpoint('storage')
        url = f'{endpoint}/{asset_id}'
        return url

    def get_listing(self, listing_id):
        endpoint = self.get_endpoint('market')
        url = f'{endpoint}/listings/{listing_id}'
        self._content_header = 'application/json'
        response = SurferAgentAdapter._http_client.get(url, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            return data
        else:
            msg = f'GET listings response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def get_listings(self):
        endpoint = self.get_endpoint('market')
        url = f'{endpoint}/listings'
        self._content_header = 'application/json'
        response = SurferAgentAdapter._http_client.get(url, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            return ResponseWrapper(response).json
        else:
            msg = f'GET listings response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def update_listing(self, listing_id, data):
        endpoint = self.get_endpoint('market')
        url = f'{endpoint}/listings/{listing_id}'
        self._content_header = 'application/json'
        response = SurferAgentAdapter._http_client.put(url, json=data, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            return True
        else:
            msg = f'PUT listings response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def upload_asset_data(self, url, asset_id, data):
        logger.debug(f'uploading data to {url}')
#        if not isinstance(dataBytes, bytes):
#            data_bytes = data.encode()

        files = {
            'file': (asset_id, io.BytesIO(data), 'application/octet-stream')
        }
#        files = {'file': (asset_id, data)}
        headers = {
            'Authorization': self._headers['Authorization'],
        }
#        self._content_header = 'application/octet-stream'
        response = SurferAgentAdapter._http_client.post(url, files=files, headers=headers)
        if response and (response.status_code == requests.codes.ok or response.status_code == requests.codes.created):
            return True
        else:
            msg = f'upload asset response failed: {response.status_code}:{response.text}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def read_metadata(self, asset_id):
        """read the metadata from a service agent using the asset_id"""

        result = None
        endpoint = self.get_endpoint('metadata')
        url = f'{endpoint}/{asset_id}'
        logger.debug(f'metadata read url {url}')
        self._content_header = 'text/plain'
        response = SurferAgentAdapter._http_client.get(url, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).data
            result = {
                'asset_id': asset_id,
                'metadata_text': data,
            }
            result['hash'] = hash_sha3_256(result['metadata_text'])
            # convert to str if bytes
            if isinstance(result['metadata_text'], bytes):
                result['metadata_text'] = data.decode('utf-8')
        else:
            logger.warning(f'metadata asset read {asset_id} response returned {response} for {url}')
        return result

    def purchase_asset(self, purchase):
        """record purchase"""
        url = self.get_endpoint('market') + '/purchases'
        logger.debug(f'market url for purchases {url}')
        self._content_header = 'application/json'
        response = SurferAgentAdapter._http_client.post(url, json=purchase, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            logger.debug(f'purchase response returned {data}')
            return data
        else:
            msg = f'purchase response failed: {response.status_code}'
            logger.error(msg)
            # NOTE surfer may return error information in additon to a 500
            if response:
                if hasattr(response, 'get_json'):
                    data = response.get_json()
                elif hasattr(response, 'json'):
                    data = response.json()
                else:
                    TypeError('cannot get the json property from response object')
                logger.error(f'purchase response returned {data}')
            raise ValueError(msg)
        return None


    def invoke(self, asset_id, params, is_async=False):
        """

        call the invoke based on the asset_id without leading 0x
        """
        endpoint = self.get_endpoint('invoke')

        if is_async:
            url = SurferAgentAdapter._invoke_api_generator.generate_url(endpoint, 'invoke_async')
        else:
            url = SurferAgentAdapter._invoke_api_generator.generate_url(endpoint, 'invoke_sync')

        url = f'{url}/{asset_id}'
        self._content_header = 'application/json'
        response = SurferAgentAdapter._http_client.post(url, json=params, headers=self._headers)
        if response and (response.status_code == requests.codes.ok or response.status_code == 201):
            data = ResponseWrapper(response).json
            logger.debug('invoke response returned: ' + str(data))
            return data
        else:
            msg = f'invoke response failed: {response.status_code} {response} for {url}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def get_job(self, job_id):
        endpoint = self.get_endpoint('invoke')
        url = SurferAgentAdapter._invoke_api_generator.generate_url(endpoint, 'jobs')
        url = f'{url}/{job_id}'
        self._content_header = 'application/json'
        response = SurferAgentAdapter._http_client.get(url, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            return data
        else:
            msg = f'GET job response failed: {response.status_code} for {url}'
            logger.error(msg)
            raise ValueError(msg)
        return None


    def get_endpoint(self, name):
        """return the endpoint based on the name of the service or service type"""
        service_type = SurferAgentAdapter.find_supported_service_type(name)
        if service_type is None:
            message = f'unknown surfer endpoint service name or type: {name}'
            logger.error(message)
            raise ValueError(message)

        endpoint = None
        if self._ddo:
            service = self._ddo.get_service(service_type)
            if not service:
                message = f'unable to find surfer endpoint service type {service_type}'
                logger.error(message)
                raise ValueError(message)
            endpoint = service.endpoints.service
        if not endpoint:
            message = f'unable to find surfer endpoint for {name} = {service_type}'
            logger.error(message)
            raise ValueError(message)
        return endpoint


    @property
    def ddo(self):
        """return the DDO stored for this agent"""
        return self._ddo

    @property
    def did(self):
        """return the DID used for this agent"""
        return self._did

    @property
    def register_name(self):
        return self._register_name

    @staticmethod
    def is_metadata_valid(asset_id, metadata_text):
        """
        validate metadata, by calcualating the hash (asset_id) and compare this to the
        given asset_id, if both are equal then the metadata is valid
        :param asset_id: asset id to compare with
        :param metadata: dict of metadata to calculate the hash ( asset_id)
        :return bool True if metadata is valid for the asset_id provided
        """
        if metadata_text:
            # the calc asset_id from the metadata should be same as this asset_id
            metadata_id = SurferAgentAdapter.get_asset_id_from_metadata(metadata_text)
            if metadata_id != asset_id:
                logger.debug(f'metdata does not match {metadata_id} != {asset_id}')
            return metadata_id == asset_id
        return False


    @staticmethod
    def get_asset_id_from_metadata(metadata_text):
        """
        return the asset_id calculated from the metadata
        :param metadata: dict of metadata to hash
        a 64 char hex string, which is the asset id
        :return 64 char hex string, with no leading '0x'
        """
        return hash_sha3_256(metadata_text)

    @staticmethod
    def set_http_client(http_client):
        """Set the http client to something other than the default `requests`"""
        SurferAgentAdapter._http_client = http_client

    @staticmethod
    def set_invoke_api_generator(invoke_api_generator):
        SurferAgentAdapter._invoke_api_generator = invoke_api_generator

    @staticmethod
    def get_authorization_token(surfer_url, surfer_username, surfer_password):
        """Get a surfer authorization token (create one if needed).
        Throws exception on error."""
        token_url = surfer_url + SURFER_BASE_URI + '/auth/token'
        response = requests.get(token_url, auth=(surfer_username, surfer_password))
        token = None
        if response.status_code == 200:
            tokens = ResponseWrapper(response).json
            if len(tokens) > 0:
                token = tokens[-1]
            else:   # need to create a token
                response = requests.post(token_url, auth=(surfer_username, surfer_password))
                if response.status_code == 200:
                    token = ResponseWrapper(response).json
                else:
                    msg = f'unable to create token, status {response.status_code}'
                    logger.error(msg)
                    raise ValueError(msg)
        else:
            msg = f'unable to get tokens, status {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        logger.debug(f'using surfer token {token}')
        return token

    @staticmethod
    def find_supported_service_type(search_name_type):
        """ return the supported service record if the name or service type is found
        else return None """
        for name, service_type in SUPPORTED_SERVICES.items():
            if service_type == search_name_type:
                return service_type
            if name == search_name_type:
                return service_type
        return None

    @property
    def _content_header(self, value):
        return self._headers['content-type']

    @_content_header.setter
    def _content_header(self, value):
        self._headers['content-type'] = value



"""
    test_06_agent_endpoint_query


    As a developer working with Ocean,
    I need to obtain service endpoints for an arbitrary Agent in the Ocean ecosystem

"""

import re

from starfish.agent import RemoteAgent
from starfish.ddo.starfish_ddo import StarfishDDO


def test_06_agent_endpoint_query(ocean, remote_agent):

    endpoint = remote_agent.get_endpoint('metadata')
    assert(re.search('/meta/data', endpoint))
    endpoint = remote_agent.get_endpoint('storage')
    assert(re.search('/assets', endpoint))
    endpoint = remote_agent.get_endpoint('invoke')
    assert(re.search('/api/v1$', endpoint))
    endpoint = remote_agent.get_endpoint('market')
    assert(re.search('/market$', endpoint))
    endpoint = remote_agent.get_endpoint('trust')
    assert(re.search('/trust$', endpoint))
    endpoint = remote_agent.get_endpoint('auth')
    assert(re.search('/auth$', endpoint))



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import shutil
from sphinx.domains.python import PythonDomain

print(os.path.abspath('../../'))
sys.path.insert(0, os.path.abspath('../../'))

# -- Project information -----------------------------------------------------

project = 'starfish-py'
copyright = 'starfish-py contributors'
author = 'starfish-py contributors'

# The full version, including alpha/beta/rc tags
release = '0.6.0'
# The short X.Y version
release_parts = release.split('.')  # a list
version = release_parts[0] + '.' + release_parts[1] + '.' + release_parts[2]

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.intersphinx',
    'sphinxcontrib.apidoc',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.graphviz',
    'sphinxcontrib.plantuml',
    'sphinx_automodapi.automodapi',
]

# apidoc settings
# See https://github.com/sphinx-contrib/apidoc
apidoc_module_dir = '../../starfish'
# apidoc_output_dir = 'api' by default, and leave it that way!
apidoc_separate_modules = True
# See https://www.sphinx-doc.org/en/master/man/sphinx-apidoc.html
apidoc_extra_args = []
apidoc_excluded_paths = [
  'ocean.py',
  'command_line',
  'constants.py',
  'exceptions.py',
  'logging.py',
  'middleware',
  'utils',
  'agent',
  'purchase',
  'listing',
  'account',
  'asset'
]

# autodoc settings
# Setting None is equivalent to giving the option name
# in the list format (i.e. it means “yes/true/on”).
# autodoc_default_options = {
#     'members': None,
#     'member-order': 'bysource',
#     'undoc-members': None,
#     'private-members': None,
#     'special-members': None,
#     'inherited-members': None,
# }

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = []

# The name of the Pygments (syntax highlig1hting) style to use.
pygments_style = 'sphinx'

highlight_language = 'python3'

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'starfishapi': ('https://dex-company.github.io/starfish-py', None)
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# For the Alabaster theme, see:
# https://alabaster.readthedocs.io/en/latest/customization.html

"""
# Colors
brand_white = '#fff'
brand_black = '#141414'
brand_pink = '#ff4092'
brand_purple = '#7b1173'
brand_grey_light = '#8b98a9'
brand_grey = '#41474e'
brand_grey_dark = '#303030'

# Fonts
font_family_base = ("'Sharp Sans Medium', -apple-system, BlinkMacSystemFont, "
                    "'Segoe UI', Helvetica, Arial, sans-serif")
font_family_title = ("'Sharp Sans Display', -apple-system, "
                     "BlinkMacSystemFont, "
                     "'Segoe UI', Helvetica, Arial, sans-serif")
font_family_monospace = ("'Fira Code', 'Fira Mono', Menlo, Monaco, Consolas, "
                         "'Courier New', monospace")

html_theme_options = {
    # Colors
    'logo': 'logo-dex.svg',
    'logo_name': True,
    'logo_text_align': 'center',
    'show_powered_by': False,
    'body_text': brand_black,
    'gray_1': brand_grey_dark,  # dark gray
    'gray_2': brand_grey_light,  # light gray
    'gray_3': brand_grey,  # medium gray
    'link': brand_pink,
    'link_hover': brand_pink,
    'pink_1': brand_pink,
    'pink_2': brand_purple,
    'sidebar_header': brand_black,
    'sidebar_link': brand_grey,
    'sidebar_list': brand_black,
    'sidebar_link_underscore': brand_white,
    'sidebar_text': brand_black,
    'code_highlight': brand_white,
    # Fonts
    'caption_font_family': font_family_base,
    'code_font_family': font_family_monospace,
    'font_family': font_family_base,
    'head_font_family': font_family_title,
}
"""

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {'sidebar.rst'}

# Delete the api/ directory of auto-generated .rst docs files
print("Removing the api/ directory via conf.py, if api/ exists.")
shutil.rmtree('api', ignore_errors=True)
print("Done removal.")

# Fix for
# https://github.com/sphinx-doc/sphinx/issues/3866
_desired_base_module = 'starfish'

class MyPythonDomain(PythonDomain):
    def find_obj(self, env, modname, classname, name, type, searchmode=0):
        """Ensures an object always resolves to the desired module if defined there."""
        orig_matches = PythonDomain.find_obj(self, env, modname, classname, name, type, searchmode)
        matches = []
        for match in orig_matches:
            match_name = match[0]
            desired_name = _desired_base_module + '.' + name.strip('.')
            if match_name == desired_name:
                matches.append(match)
                break
        if matches:
            return matches
        else:
            return orig_matches


def setup(sphinx):
    """Use MyPythonDomain in place of PythonDomain"""
    sphinx.override_domain(MyPythonDomain)



"""
    Asset class to handle core imutable asset and it's metadata


"""

import json
from abc import ABC
from starfish.utils.did import did_to_asset_id

class AssetBase(ABC):
    """

    :param str metadata_text: metadata text for the asset
    :param did: Octional did of the asset, if the asset is new then the did will be None.
    :type did: None or str

    """
    def __init__(self, metadata, did=None, metadata_text=None):
        """
        init an asset class
        """

        if not isinstance(metadata, dict):
            raise ValueError('metadata must be a dict')

        self._metadata = metadata
        self._metadata_text = metadata_text

        if self._metadata_text is None:
            self._metadata_text = json.dumps(metadata)

        if 'name' not in metadata:
            raise ValueError('metadata must contain a metadata name')

        if 'type' not in metadata:
            raise ValueError('metadata must contain a metadata type')

        self._did = did
        super().__init__()


    def set_did(self, did):
        """
        This method makes the object immutable.
        So maybe a solution is that we have a 'copy' and
        set the did in the __init__ of the new class, to return a mutable copy of the
        same asset object.
        """
        self._did = did

    def is_asset_type(self, type_name):
        """

        Returns if this metadata has the correct type

        :param str type_name: name of the asset type stored in the metadata

        :return: True if the metadata type is equal to type_name
        :type: boolean

        """
        asset_type = AssetBase.get_asset_type(self._metadata)
        return asset_type == type_name

    @property
    def did(self):
        """
        :return: the asset did
        :type: str
        """
        return self._did

    @property
    def metadata(self):
        """
        :return: The metadata for this asset
        :type: dict
        """
        return self._metadata

    @property
    def metadata_text(self):
        """
        :return: The metadata for this asset
        :type: dict
        """
        return self._metadata_text

    @property
    def asset_id(self):
        if self._did:
            return did_to_asset_id(self._did)
        return None

    @property
    def name(self):
        return self._metadata['name']

    @property
    def type_name(self):
        return self._metadata['type']

    @property
    def is_bundle(self):
        """

        Return True if this asset is a bundle asset and can contain sub assets ( Asset Bundle )

        :return: True if sub assets can be held within this asset
        :type: boolean

        """
        return False


    @staticmethod
    def get_asset_type(metadata):
        asset_type = ''
        if isinstance(metadata, dict):
            if 'type' in metadata:
                asset_type = metadata['type']
            else:
                # if from squid then it's always a bundle
                if 'base' in metadata:
                    asset_type = 'bundle'
        return asset_type

    @staticmethod
    def generateMetadata(name, asset_type, metadata=None):
        if metadata is None:
            metadata = {}
        if not isinstance(metadata, dict):
            raise ValueError('The metadata has to be a dict')

        metadata['name'] = name
        metadata['type'] = asset_type
        return metadata




"""
Operation class
"""
from abc import ABC, abstractmethod


class OperationBase(ABC):

    def __init__(self, agent,did):
        """init the the Operation with the agent instance"""
        self._agent = agent
        self._did=did
        super().__init__()

    @abstractmethod
    def invoke(self, **kwargs):
        """

        """
        pass



"""
    test_14_view_asset_listing

    As a developer working with Ocean, I need a way to view an asset available
    for puchase and any associated terms / service agreements that may
    be purchased

"""

import logging
import json

from starfish.asset import DataAsset


def test_14_view_asset_listing(resources, remote_agent):
    test_data = "Test listing searching by listing id"
    asset = DataAsset.create('TestAsset', test_data)
    listing = remote_agent.register_asset(asset, resources.listing_data)
    listing.set_published(True)
    logging.debug("view_asset_listing for listing_id: " + listing.listing_id)
    listing2 = remote_agent.get_listing(listing.listing_id)
    assert(listing2.listing_id == listing.listing_id)
    # view all listings
    listings = remote_agent.get_listings()
    logging.debug("listings: " + str(listings))
    assert(listings)
    assert(isinstance(listings,list))
    assert(len(listings) > 0)
    # found = False
    # FIXME: the remote agent call to get all listings does NOT return this listing!?
    found = True
    for listing in listings:
        logging.debug("checking: " + str(listing.listing_id) + " == " + str(listing.listing_id == listing2.listing_id))
        if listing.listing_id == listing2.listing_id:
            found = True
            break
    assert(found)



"""

Config Class to handle config data and config files

"""

import configparser
import logging
import os
import os.path
import re
import tempfile

from starfish import logger

CONFIG_SECTION_NAME = 'starfish-py'

CONFIG_DEFAULT = """
[starfish-py]
keeper_url = http://localhost:8545
contract_path = artifacts

secret_store_url = http://localhost:8010
parity_url = http://localhost:9545

aquarius_url = http://localhost:5000
brizo_url = http://localhost:8030

storage_path = squid_py.db
download_path = consume_downloads

agent_store_did =
agent_store_auth =

gas_limit = 300000

"""


class Config(configparser.ConfigParser):
    """
    Create a new config instance, using a config file, dictionary of values or argument list (kwargs)

    :param filename: Filename of the config file to load.
    :type filename: str or None
    :param contracts_path: path to the contract files ( artifacts ).
    :type contracts_path: str or None
    :param keeper_url: url to the keeper node ( http://localhost:8545 ).
    :type keeper_url: str or None
    :param secret_store_url: url to the secret store node ( http://localhost:12001 ).
    :type secret_store_url: str or None
    :param parity_url: url to the parity node ( http://localhost:8545 ).
    :type parity_url: str or None
    :param aquarius_url: url of the Aquarius metadata service ( http://localhost:5000 ).
    :type aquarius_url: str or None
    :param brizo_url: url of the Brizo consumer service (http://localhost:8030 ).
    :type brizo_url: str or None
    :param storage_path: Path to save temporary storage of assets purchased and consumed ( squid_py.db ).
    :type storage_path: str or None
    :param download_path: Path to save the consumed assets too. ( consume_downloads ).
    :type download_path: str or None
    :param agent_store_did: DID of the agent metadata service.
    :type agent_store_did: str or None
    :param agent_store_auth: Authorziation text to access the metadat service.
    :type agent_store_auth: str or None
    :param gas_limit: The amount of gas you are willing to spend on each block chain transaction ( 30000 ).
    :type gas_limit: int or string

    For example::

        # only set the contarcts path and keeper url values
        ocean = Ocean(contracts_path='artifacts', keeper_url='http://localhost:8080')

    The default config values as defined as an `ini` file are defined as follows:

    .. literalinclude:: ../../../starfish/config.py
        :start-after: CONFIG_DEFAULT
        :end-before: "
    """

    def __init__(self, filename=None, **kwargs):
        """
        The config class

        """
        configparser.ConfigParser.__init__(self)

        self.read_string(CONFIG_DEFAULT)
        self._section_name = CONFIG_SECTION_NAME
        values = {}

        if filename:
            if isinstance(filename, str):
                if not os.path.exists(filename):
                    logging.error(f'Config file not found: "{filename}"')
                logging.debug(f'loading config file {filename}')
                with open(filename) as file_handle:
                    text = file_handle.read()
                    self.read_string(text)
                    values[self._section_name] = kwargs
            elif isinstance(filename, dict):
                logger.debug('loading config from a dict')
                values[self._section_name] = filename
            elif isinstance(filename, Config):
                logger.debug('loading config from another config object')
                for section_name in filename.items():
                    values[section_name[0]] = {}
                    for name, value in filename.items(section_name[0]):
                        values[section_name[0]][name] = value
            else:
                raise TypeError('Invalid type of data passed, can only be a filename or a dict of values')
        else:
            if kwargs:
                logger.debug('loading config from a kwargs')
                values[self._section_name] = kwargs

        if values:
            logger.debug(f'loading values {kwargs}')
            self.read_dict(values)

        self._read_environ()

    def _read_environ(self):
        """

        Read the environment variables and replace them with the config values

        """
        defaults = configparser.ConfigParser()
        defaults.read_string(CONFIG_DEFAULT)
        for name, value in defaults.items(CONFIG_SECTION_NAME):
            value = os.environ.get(re.sub(r'[^\w]+', '_', name).upper())
            if value is not None:
                logger.debug(f'setting environ {name} = {value}')
                self.set(self._section_name, name, value)

    @property
    def as_squid_file(self):
        """

        For compatibility generate a temporary config file, and pass back the filename.
        The config values must conform to the current version of squid-py

        :return: A temporary filename that can be used by squid.
        :type: str

        """
        squid = configparser.ConfigParser()
        values = self.as_squid_dict
        squid.read_dict(values)
        logger.debug(f'squid config values {values}')
        temp_handle = tempfile.mkstemp('_squid.conf', text=True)
        os.close(temp_handle[0])
        filename = temp_handle[1]
        with open(filename, 'w') as file_handle:
            squid.write(file_handle)
        return filename

    def as_squid_dict(self, options = None):
        """

        Return a set of config values, so that squid can read.

        :param options: optional values to add to the dict to send to squid
        :type options: dict or None


        :return: a dict that is compatiable with the current supported version of squid-py.
        :type: dict

        """
        data = {
            'keeper-contracts': {
                'keeper.url': self.keeper_url,
                'keeper.path': self.contract_path,
                'secret_store.url': self.secret_store_url,
                'parity.url': self.parity_url,
            },
            'resources': {
                'aquarius.url': self.aquarius_url,
                'brizo.url': self.brizo_url,
                'storage.path': self.storage_path,
                'downloads.path': self.download_path,
            }
        }
        if isinstance(options, dict):
            if 'parity_address' in options:
                data['keeper-contracts']['parity.address'] = options['parity_address']
            if 'parity_password' in options:
                data['keeper-contracts']['parity.password'] = options['parity_password']

        return data

    @property
    def contract_path(self):
        """
        :return: the contract path value.
        :type: str
        """
        return self.get(self._section_name, 'contract_path')

    @property
    def storage_path(self):
        """
        :return: the storage path.
        :type: str
        """
        return self.get(self._section_name, 'storage_path')

    @property
    def download_path(self):
        """
        :return: the download path.
        :type: str
        """
        return self.get(self._section_name, 'download_path')

    @property
    def keeper_url(self):
        """
        :return: the ocean url or ethereum node url.
        :type: str
        """
        return self.get(self._section_name, 'keeper_url')

    @property
    def gas_limit(self):
        """
        :return: the default gas limit
        :type: int
        """
        return int(self.get(self._section_name, 'gas_limit'))

    @property
    def aquarius_url(self):
        """
        :return: the aquarius server URL.
        :type: str
        """
        return self.get(self._section_name, 'aquarius_url')

    @property
    def brizo_url(self):
        """
        :return: the URL of the brizo server.
        :type: str
        """
        return self.get(self._section_name, 'brizo_url')

    @property
    def secret_store_url(self):
        """
        :return: the secret store URL.
        :type: str
        """
        return self.get(self._section_name, 'secret_store_url')

    @property
    def parity_url(self):
        """
        :return: the parity URL.
        :type: str
        """
        return self.get(self._section_name, 'parity_url')

    @property
    def agent_store_did(self):
        """
        :return: the storage agent's DID.
        :type: str
        """
        return self.get(self._section_name, 'agent_store_did')

    @property
    def agent_store_auth(self):
        """
        :return: the storage agent's authorization.
        :type: str
        """
        return self.get(self._section_name, 'agent_store_auth')



"""
    test_12_listing_publish

    As a developer working with an Ocean marketplace,
    I need a way to unpublish my asset (i.e. remove relevant listings) from a marketplace

"""

import secrets

from starfish.asset import DataAsset


def test_12_listing_publish(resources, remote_agent):
    test_data = secrets.token_hex(1024)
    asset = DataAsset.create('TestAsset', test_data)
    listing = remote_agent.register_asset(asset, resources.listing_data)
    assert(listing)
    assert(listing.asset)

    assert(not listing.is_published)

    listing.set_published(True)
    assert(listing.is_published)

    assert(remote_agent.update_listing(listing))

    read_listing = remote_agent.get_listing(listing.listing_id)
    assert(read_listing)
    assert(read_listing.is_published)



"""
    test_03_agent_identity


    As a developer working with Ocean,
    I need a stable identifier (Agent ID) for an arbitrary Agent in the Ocean ecosystem
    This test class with validate the DID and DDO format and and their data

"""

import secrets
from starfish.agent import RemoteAgent
from starfish.utils.did import did_parse
from starfish.ddo.starfish_ddo import StarfishDDO


def find_remote_service(find_service_type):
    for service_name, service_type in RemoteAgent.service_types.items():
        if service_type == find_service_type:
            return service_type
    return None

def test_03_agent_did(remote_agent):
    did = remote_agent.did
    assert(did)
    data = did_parse(did)
    assert(data)
    assert(data['method'] == 'dep')
    assert(data['id_hex'] == '0x' + data['id'])
    assert(data['path'] is None)
    assert(data['fragment'] is None)

def test_03_agent_ddo(remote_agent):
    ddo = remote_agent.ddo
    assert(ddo)
    assert(isinstance(ddo, StarfishDDO))
    for service in ddo.services:
        service_type = find_remote_service(service.type)
        assert(service_type)
        endpoint = remote_agent.get_endpoint(service_type)
        assert(endpoint)
        assert(endpoint == service.endpoints.service)


def test_03_agent_get_endpoints(remote_agent):
    assert(remote_agent.get_endpoint('metadata'))
    assert(remote_agent.get_endpoint('storage'))
    assert(remote_agent.get_endpoint('invoke'))
    assert(remote_agent.get_endpoint('market'))
    assert(remote_agent.get_endpoint('trust'))
    assert(remote_agent.get_endpoint('auth'))




"""
    Base Purchase class
"""
from abc import ABC, abstractmethod


class PurchaseBase(ABC):

    def __init__(self, agent, listing, purchase_id, account):
        """init the the Purchase Object Base with the agent instance"""
        self._agent = agent
        self._listing = listing
        self._purchase_id = purchase_id
        self._account = account
        super().__init__()

    @property
    def agent(self):
        """
        :return: Agent object
        :type: :class:`.Agent`
        """
        return self._agent

    @property
    def listing(self):
        """
        :return: Listing object
        :type: :class:`.Listing'`
        """
        return self._listing

    @property
    def purchase_id(self):
        """
        :return: purchase id
        :type: str
        """
        return self._purchase_id

    @property
    def account(self):
        """
        :return: account that is doing the purchase
        :type: :class:`.Account`
        """
        return self._account

    @property
    @abstractmethod
    def is_purchased(self):
        """
        :return: True if this asset is a purchased asset.
        :type: boolean
        """
        pass

    @property
    @abstractmethod
    def is_purchase_valid(self):
        """
        :return: True if this asset is a valid purchase.
        :type: boolean
        """
        pass



import secrets

class UnitTestAccount():

    def __init__(self, password=None, keyfile=None):
        self.test_address = '0x' + secrets.token_hex(20)
        self.test_password = secrets.token_hex(32)
        self.test_keyfile = secrets.token_hex(32)
        if password:
            self.test_password = password
        if keyfile:
            self.test_keyfile = keyfile
        self.test_ether = 1 + secrets.randbelow(100)
        self.test_tokens = 5 + secrets.randbelow(100)
        self._agent_adapter = None

    @property
    def as_tuple(self):
        return (self.test_address, self.test_password, self.test_keyfile)

    @property
    def as_dict(self):
        return {'address': self.test_address, 'password': self.test_password, 'keyfile': self.test_keyfile}

    @property
    def as_list(self):
        return [self.test_address, self.test_password, self.test_keyfile]

    @property
    def address(self):
        return self.test_address

    @property
    def agent_adapter(self):
        return self._agent_adapter

    @agent_adapter.setter
    def agent_adapter(self, agent_adapter):
        self._agent_adapter = agent_adapter




"""
    test_02_asset_identity


    As a developer working with Ocean,
    I need a stable identifier (Asset ID) for an arbitrary asset in the Ocean ecosystem

"""

import secrets
from starfish.asset import DataAsset
from starfish.agent import RemoteAgent

# TEST_TEXT = "Django raspberrypi diversity goat python. Cython raspberrypi community dict import object coroutine arduino. Python raspberrypi beautiful pypy gevent pycon raspberrypi rocksdahouse method. Return integration mercurial pypi dunder. Reduce integration itertools test import. Future raspberrypi exception list."

def test_02_asset_register(ocean, resources, remote_agent):
    testData = secrets.token_hex(1024)
    asset1 = DataAsset.create('TestAsset1', testData)
    asset2 = DataAsset.create('TestAsset2', testData)
    listing = remote_agent.register_asset(asset2, resources.listing_data)
    assert(not listing is None)
    assert(64==len(listing.listing_id))
    assert(listing.asset.data == asset2.data)
    assert(listing.asset.asset_id == asset2.asset_id)


def test_02_asset_upload(ocean, resources, remote_agent):
    testData = secrets.token_hex(1024)
    asset = DataAsset.create('TestAsset', testData)
    listing = remote_agent.register_asset(asset, resources.listing_data)
    remote_agent.upload_asset(listing.asset)



"""koi client ."""
import sys,traceback
import json
import logging
import os

import requests
from tqdm import tqdm

from eth_utils import add_0x_prefix
from squid_py.agreements.service_agreement import ServiceAgreement
from squid_py.exceptions import OceanInitializeServiceAgreementError

from squid_py.brizo.brizo import Brizo
logger = logging.getLogger(__name__)


class KoiMock:
    """

    The main functions available are:
    - initialize_service_agreement
    - consume_service
    - run_compute_service (not implemented yet)

    """
    _http_client = requests

    @staticmethod
    def set_http_client(http_client):
        """Set the http client to something other than the default `requests`"""
        KoiMock._http_client = http_client

    publisher_account = None
    ocean_instance = None
    
    def __init__(self, ocean_instance=None, account=None):
        self.ocean_instance = ocean_instance
        if not self.ocean_instance:
            self.ocean_instance = KoiMock.ocean_instance

        self.account = account
        if not self.account:
            self.account = KoiMock.publisher_account

    def initialize_service_agreement(self,did, agreement_id, service_definition_id, signature,
                                     account_address,
                                     purchase_endpoint):
        """
        Send a request to the service provider (purchase_endpoint) to initialize the service
        agreement for the asset identified by `did`.

        :param did: str -- id of the asset includes the `did:op:` prefix
        :param agreement_id: hex str
        :param service_definition_id: str -- identifier of the service inside the asset DDO
        :param signature: hex str -- signed agreement hash
        :param account_address: hex str -- ethereum address of the consumer signing this agreement
        :param purchase_endpoint: str -- url of the service provider
        :return:
        """
        
        logger.info(f' Koi client mocking request to server')

        receipt = self.ocean_instance.agreements.create(did, 
                service_definition_id, 
                agreement_id, 
                signature, 
                account_address, 
                self.account
        )
        return True
 
    @staticmethod
    def invoke_service(service_agreement_id, service_endpoint, account_address, invoke_payload):
        """
        Call the Koi endpoint to get access to the different files that form the asset.

        :param service_agreement_id: Service Agreement Id, str
        :param service_endpoint: Url to consume, str
        :param account_address: ethereum address of the consumer signing this agreement, hex-str
        :param files: List containing the files to be consumed, list
        :param destination_folder: Path, str
        :return:
        """
        logger.info(f'invoke endpoint with this url: {service_endpoint}')
            
        return invoke_payload['params'] 

    @staticmethod
    def get_brizo_url(config):
        """
        Return the Brizo component url.

        :param config: Config
        :return: Url, str
        """
        brizo_url = 'http://localhost:8031'
        if config.has_option('resources', 'brizo.url'):
            brizo_url = config.get('resources', 'brizo.url') or brizo_url

        brizo_path = '/api/v1/brizo'
        return f'{brizo_url}{brizo_path}'

    @staticmethod
    def get_purchase_endpoint(config):
        """
        Return the endpoint to purchase the asset.

        :param config:Config
        :return: Url, str
        """
        return f'{Brizo.get_brizo_url(config)}/services/access/initialize'

    @staticmethod
    def get_service_endpoint(config):
        """
        Return the url to consume the asset.

        :param config: Config
        :return: Url, str
        """
        return f'{Brizo.get_brizo_url(config)}/services/invoke'



""""

Test examples

"""


def test_register_asset():
    from examples.register_asset_squid import main
    main()


def test_register_upload_asset_remote():
    from examples.register_upload_asset_remote_agent import main
    main()



"""
    Test ocean class

"""

import datetime
import pathlib
import json
import logging
import time
import secrets
import pytest

from web3 import Web3

from starfish import Ocean
from starfish.agent import SquidAgent
from starfish.asset import (
    DataAsset,
    RemoteDataAsset,
)

from starfish.exceptions import (
    StarfishAssetNotFound,
    StarfishPurchaseError
)

from ocean_utils.agreements.service_factory import ServiceDescriptor
from ocean_utils.utils.utilities import generate_new_id

from squid_py.brizo.brizo_provider import BrizoProvider
from ocean_utils.aquarius.aquarius import Aquarius
from ocean_utils.ddo.ddo import DDO

logger = logging.getLogger('test.core.test_squid')

def _register_asset_for_sale(agent, resources, account):

    asset = DataAsset.create_from_file('TestAsset', resources.asset_file)
    listing = agent.register_asset(asset, resources.listing_data, account=account)
    assert(listing)
    assert(listing.asset.did)
    return(listing)


def test_asset_file_register(ocean, config, resources, publisher_account):

    agent = SquidAgent(ocean, config.squid_config)
    assert(agent)

    asset = DataAsset.create_from_file('TestAsset', resources.asset_file)
    listing = agent.register_asset(asset, resources.listing_data, publisher_account)
    assert(listing)


def test_asset_remote_register(ocean, config, resources, publisher_account):

    agent = SquidAgent(ocean, config.squid_config)
    assert(agent)

    asset = RemoteDataAsset.create_with_url('TestAsset', resources.asset_remote)

    listing = agent.register_asset(asset, resources.listing_data, publisher_account)
    assert(listing)


def is_purchase_allowed(did, agreement_id, publish_account, conusmer_account):

    logger.debug(f'is purchase allowed {did}, {agreement_id}, {publish_account}, {conusmer_account}')
    return True

def test_asset_purchase(ocean, config, resources, publisher_account, purchaser_account):

    agent = SquidAgent(ocean, config.squid_config)
    assert(agent)

    publisher_account.request_tokens(20)

    listing = _register_asset_for_sale(agent, resources, publisher_account)
    assert(listing)
    assert(publisher_account)

    listing_did = listing.asset.did
    # start to test getting the asset from storage
    listing = agent.get_listing(listing_did)
    assert(listing)
    assert(listing.asset.did == listing_did)



    logger.info(f'purchase_account {purchaser_account.ocean_balance}')

    # test is_purchased for an account only
    assert(not listing.is_purchased(purchaser_account))

    purchaser_account.request_tokens(10)


    time.sleep(1)
    logger.info(f'purchase_account after token request {purchaser_account.ocean_balance}')


#    agent.start_agreement_events_monitor(publisher_account, is_purchase_allowed)
    agent.start_agreement_events_monitor(publisher_account, None)

    # test purchase an asset
    purchase_asset = listing.purchase(purchaser_account)
    assert(purchase_asset)

    assert(not purchase_asset.is_completed)

    error_message = purchase_asset.wait_for_completion()
    assert(error_message == True)

    assert(purchase_asset.is_completed)

    # assert Web3.toHex(event.args['_agreementId']) == agreement_id
    # assert len(os.listdir(consumer_ocean_instance.config.downloads_path)) == downloads_path_elements + 1

    # This test does not work with the current barge

    assert(purchase_asset.is_purchased)
    assert(purchase_asset.is_purchase_valid)


    # test getting a list of purchase ids from a listing
    purchase_ids = listing.get_purchase_ids
    assert(purchase_ids)
    assert(len(purchase_ids) == 1)
    assert(purchase_ids[0] == purchase_asset.purchase_id)

    # test is_purchased for an account only
    assert(listing.is_purchased(purchaser_account))

    remote_asset = purchase_asset.consume_asset
    assert(remote_asset)


def test_search_listing(ocean, config, resources, publisher_account):


    agent = SquidAgent(ocean, config.squid_config)


    listing = _register_asset_for_sale(agent, resources, publisher_account)
    assert listing
    assert publisher_account

    # choose a word from the description field
    text = resources.listing_data['author']
    words = text.split(' ')
    word = words[0]

    # should return at least 1 or more assets
    logger.info(f'search word is {word}')
    search_result = agent.search_listings(word)
    assert(search_result)

    assert(len(search_result) > 1)

def test_get_listing(ocean, config, resources, publisher_account, purchaser_account):
    agent = SquidAgent(ocean, config.squid_config)

    listing = _register_asset_for_sale(agent, resources, publisher_account)
    assert listing
    assert publisher_account

    found_listing = agent.get_listing(listing.listing_id)
    assert(found_listing)

    # now test for an asset in aquarius but not on the block chain network

    aquarius = Aquarius(config.squid_config['aquarius_url'])
    ddo = aquarius.get_asset_ddo(listing.listing_id)
    dummy_listing_id = f'did:op:{generate_new_id()}'
    ddo_dict = ddo.as_dictionary()
    ddo_dict['id'] = dummy_listing_id
    ddo_dict['publicKey'][0]['id'] = dummy_listing_id
    ddo_dict['authentication'][0]['publicKey'] = dummy_listing_id
    ddo_dict['created'] = ddo_dict['created'] + 'Z'
    ddo_dict['service'][0]['metadata']['base']['dateCreated'] = ddo_dict['service'][0]['metadata']['base']['dateCreated'] + 'Z'
    # fix squid index bug in service list ?
    index = 0
    for service in ddo_dict['service']:
        ddo_dict['service'][index]['serviceDefinitionId'] = index

    dummy_ddo = DDO(dictionary=ddo_dict)
    aquarius.publish_asset_ddo(dummy_ddo)

    ddo = aquarius.get_asset_ddo(dummy_listing_id)
    assert(ddo.did == dummy_listing_id)

    # the final test, the asset did is in aquarius, but the did -> url is not on the block chain
    with pytest.raises(StarfishAssetNotFound):
        dummy_listing = agent.get_listing(dummy_listing_id)

    # now try to purchase an invalid asset with no block chain DID

    listing_list = agent.search_listings({'tags': ['starfish']})
    assert(listing_list)
    for listing in listing_list:
        assert(listing)
        if listing.listing_id == dummy_listing_id:
            with pytest.raises(StarfishAssetNotFound):
                listing.purchase(purchaser_account)

def test_insufficient_funds(ocean, config, resources, publisher_account, purchaser_account):

    agent = SquidAgent(ocean, config.squid_config)
    assert(agent)

    asset = DataAsset.create_from_file('TestAsset', resources.asset_file)
    listing_data = resources.listing_data.copy()
    listing_data['price'] = 83454.2345
    listing = agent.register_asset(asset, listing_data, publisher_account)
    assert(listing)

    listing_did = listing.asset.did
    # start to test getting the asset from storage
    listing = agent.get_listing(listing_did)

    with pytest.raises(StarfishPurchaseError):
        purchase_asset = listing.purchase(purchaser_account)



"""

Ocean class to access the Ocean eco system.

"""
import logging

from starfish.utils.artifacts import (
    find_contract_path,
    is_contract_type_exists,
)

from web3 import (
    Web3,
    HTTPProvider
)

from starfish.account import Account
from starfish.middleware.squid_agent_adapter import SquidAgentAdapter
from squid_py.ocean.keeper import SquidKeeper


logger = logging.getLogger('starfish.ocean')

class Ocean():
    """
    .. _Asset class: asset.html
    .. _Config class: config.html

    The Ocean class connects to the ocean network.

    For example to use this class you can do the following: ::

        from starfish import Ocean

        my_config = {
            'contracts_path': 'artifacts',
            'keeper_url': 'http://localhost:8545',
            'gas_limit': 1000,
        }
        ocean = Ocean(my_config)

    or you can do the following setup: ::

        from starfish.ocean import Ocean
        ocean = Ocean(keeper_url='http://localhost:8545', contracts_path= artifacts, gas_limit=1000)

    or with no block chain network: ::

        from starfish.ocean import Ocean
        ocean = Ocean()



    You can provide these parameters in a dictionary or as individual parameters.

    :param network: name of the network to connect too. This can be 'development', 'nile' or 'kovan'
    :param keeper_url: url to the keeper node ( http://localhost:8545 ).
    :type keeper_url: str or None
    :param contracts_path: path to the contract files ( artifacts ).
        If not prodived then the network name will be used to search for the correct artifact files.
    :type contracts_path: str or None
    :param gas_limit: The amount of gas you are willing to spend on each block chain transaction ( 0 ).
    :type gas_limit: int or string

    """

    def __init__(self, *args, **kwargs):
        """
        .. :class: starfish.Ocean

        init the basic Ocean class for the connection and contract info

        """
        self.__web3 = None
        self.__squid_agent_adapter = None

        if args and isinstance(args[0], dict):
            kwargs = args[0]

        self._keeper_url = kwargs.get('keeper_url', None)
        self._network_name = kwargs.get('network', None)
        self._contracts_path = kwargs.get('contracts_path', None)
        self._gas_limit = kwargs.get('gas_limit', 0)
        self.__squid_agent_adapter_class = kwargs.get('squid_agent_adapter_class', None)

        if self._keeper_url and kwargs.get('connect', True):
            self.connect()

    def connect(self, keeper_url=None, network_name=None, contracts_path=None):
        """
        Normally you do not need to call this, since the ocean class will connect automatically
        using the provided keeper url and contracts path
        """
        if keeper_url:
            self._keeper_url = keeper_url
        if network_name:
            self._network_name = network_name
        if contracts_path:
            self._contracts_path = contracts_path

        if self._keeper_url:
            self.__web3 = Web3(HTTPProvider(self._keeper_url))
            # set the default squid adapter class if not set already
            if self.__squid_agent_adapter_class is None:
                self.__squid_agent_adapter_class = SquidAgentAdapter
            if not self._network_name:
                self._network_name = self._get_network_name()

        # check to see if the contracts path actually contain contracts for this network
        if self._contracts_path and not is_contract_type_exists(self._network_name, self._contracts_path):
            # if not then find the correct contracts path
            self._contracts_path = find_contract_path(self._network_name)
            logger.debug(f'Changing contracts path to {self._contracts_path}')

        # if no contracts path then search for the contract for this network
        if self._contracts_path is None and self._network_name:
            self._contracts_path = find_contract_path(self._network_name)

        logger.debug(f'network: {self._network_name} contracts_path: {self._contracts_path}')


    def register_did(self, did, ddo, account):
        """

        Register this agent service with a DDO on the block chain.

        :param did: DID to use to register for this ddo.
        :type did: str or None
        :param ddo: DDO to save for the registration. This is a JSON string of a DDO
        :type string: JSON string of a DDO
        :param account: account to use as the owner of the registration.
        :type account: :class:`.Account`
        :return: the receipt of the block chain transaction
        :type: string

        For example::

            # register the public remote agent on the block chain
            receipt = ocean.register_did(did, ddo.as_text(), register_account)

        TODO: Need to split this up into two calls, one to add, other to update
        """

        if not isinstance(account, Account):
            raise TypeError('You need to pass an Account object')

        if not account.is_valid:
            raise ValueError('You must pass a valid account')

        if not isinstance(ddo, str):
            raise TypeError('You need to pass a DDO as a string')

        # call the squid adapter to do the actual registration writing the ddo to the block chain
        adapter = self.get_squid_agent_adapter()
        if adapter:
            return adapter.register_ddo(did, ddo, account.agent_adapter_account)
        return None

    def resolve_did(self, did):
        """

        Return the resolved did written on the block chain
        if no value found then return None

        :param str did: did to resolve
        :return: ddo or url as a string
        :type: string

        """

        adapter = self.get_squid_agent_adapter()
        if adapter:
            return adapter.resolve_did(did)
        return None

    def search_operations(self, text, limit=10):
        """

        Search the off chain storage for an operation that matches 'text'

        :param str text: Search for 'text' in metadata.
        :param int limit: Limit the result. If not provided, default is 10
        :return: a list of where each object is a 2-tuple (service provider did, operation did)
        :type: list of 2-tuple strings

        For example: ::

            # return the first 10 records in the search for operations that do model training
            #
            my_result = ocean.search_operations('model_training')
        """
        ## To be implemented
        return []

    def load_account(self, address, password=None, keyfile=None, agent_adapter=None):
        """
        Get an account object based on it's address, password and keyfile.
        This returns the account object, of the local account.

        :param address: address of the account, if dict then use the fields, `address`, `password` and 'keyfile'.
        :type address: str, list or dict
        :param str password: password to access the keyfile for this account
        :param str keyfile: JSON filename of the account key file.
        :param agent_adapter: Optional adapter class to use for this account object. Usually it's the :class:`SquidAdapter`

        :return: return the :class:`Account` object.
        :type: :class:`Account` or None

        >>> account = ocean.load_account('0x00bd138abd70e2f00903268f3db08f2d25677c9e', 'password', 'key_file_test.json')
        """
        account = Account(self, address, password, keyfile, agent_adapter)
        return account

    @property
    def _web3(self):
        """return the web3 instance"""
        return self.__web3

    @property
    def keeper_url(self):
        return self._keeper_url

    @property
    def contracts_path(self):
        return self._contracts_path

    @property
    def gas_limit(self):
        return self._gas_limit

    @property
    def network_name(self):
        return self._network_name

    @property
    def is_connected(self):
        return not self.__web3

    def _get_network_name(self):
        network_id = int(self.__web3.version.network)
        return SquidKeeper.get_network_name(network_id)

    def get_squid_agent_adapter(self, options=None):
        if self.__squid_agent_adapter_class:
            self.__squid_agent_adapter = self.__squid_agent_adapter_class(self, options)
            return self.__squid_agent_adapter
        return None



"""
    test_04_agent_registration


    As a developer building or managing an Ocean Agent,
    I need to be able to register my Agent on the network and obtain an Agent ID

"""

from starfish.agent import RemoteAgent
from starfish.ddo.starfish_ddo import StarfishDDO



def test_04_agent_register_and_resolve(ocean, config, publisher_account):

    ddo = RemoteAgent.generate_ddo(config.remote_agent_url)
    options = {
        'url': config.remote_agent_url,
        'username': config.remote_agent_username,
        'password': config.remote_agent_password
    }

    register_account = publisher_account

    did = ddo.did
    assert(ocean.register_did(ddo.did, ddo.as_text(), register_account))
    found_ddo = StarfishDDO(json_text=ocean.resolve_did(did))
    assert(found_ddo.as_text() == ddo.as_text())

    resolved_agent = RemoteAgent(ocean, did=did)
    assert(resolved_agent)
    assert(resolved_agent.ddo)
    assert(resolved_agent.ddo.as_text() == ddo.as_text())



"""

Surfer Agent class to provide basic functionality for Ocean Agents

In starfish-java, this is named as `RemoteAgent`

"""
import time

from eth_utils import remove_0x_prefix

from starfish.agent import AgentBase
from starfish.agent.services import Services


from starfish.asset import (
    DataAsset,
    OperationAsset,
    create_asset_from_metadata_text,
    is_asset_hash_valid,
)
from starfish.middleware.surfer_agent_adapter import SurferAgentAdapter, SUPPORTED_SERVICES
from starfish.middleware.squid_agent_adapter import SquidAgentAdapter
from starfish.utils.did import (
    did_parse,
    did_generate_random,
)
from starfish.listing import Listing
from starfish.job import Job
from starfish.ddo.starfish_ddo import StarfishDDO
from starfish.exceptions import StarfishAssetInvalid

class RemoteAgent(AgentBase):
    """

    Remote Agent class allows to register, list, purchase and consume assets.

    :param ocean: Ocean object that is being used.
    :type ocean: :class:`.Ocean`
    :param did: Optional did of the Surfer agent

    :param ddo: Optional ddo of the remote agent, if not provided the agent
        will automatically get the DDO from the network based on the DID.

    :param options: Optional options, only `authorization` is used to access the
        Surfer server.

    """
    service_types = SUPPORTED_SERVICES


    def __init__(self, ocean, did=None, ddo=None, options=None):
        """init a standard ocean object"""
        AgentBase.__init__(self, ocean)
        self._did = None
        self._ddo = None

        if options is None:
            options = {}

        # set the DID
        if did is None or isinstance(did, str):
            self._did = did
        else:
            raise ValueError('did must be a string or None')

        # set the DDO
        if isinstance(ddo, dict):
            self._ddo = StarfishDDO(dictionary=ddo)
        elif isinstance(ddo, str):
            self._ddo = StarfishDDO(json_text=ddo)
        elif isinstance(ddo, StarfishDDO):
            self._ddo = ddo
        elif ddo is None:
            if self._did:
                self._ddo = self._resolve_ddo_from_did(self._did)
        else:
            raise ValueError('ddo can be one of the following: None, StarfishDDO object or type dict')

        # incase the user just sends a ddo without a did
        if self._did is None and self._ddo:
            self._did = self._ddo.did


        # if DID and no DDO then try to load in the registered DDO, using squid
        if self._did and not self._ddo:
            squid_adapter = SquidAgentAdapter(ocean)
            ddo_text = squid_adapter.resolve_did_to_ddo(self._did)
            if not ddo_text:
                raise ValueError(f'cannot find registered agent at {self.did}')
            self._ddo = StarfishDDO(json_text=ddo_text)

        if self._did is None and self._ddo:
            self._did = self._ddo.did

        self._authorization = options.get('authorization')
        if self._authorization is None and 'url' in options and 'username' in options:
            # if no authorization, then we may need to create one
            self._authorization = SurferAgentAdapter.get_authorization_token(
                options['url'],
                options['username'],
                options.get('password', '')
            )

    def register_asset(self, asset, listing_data, account=None):
        """

        Register an asset with Surfer

        :param asset: asset object to register
        :type asset: :class:`.DataAsset` object to register
        :param dict listing_data:  Listing inforamiton to give for this asset
        :param account: This is not used for this agent, so for compatibility it is left in
        :type account: :class:`.Account` object to use for registration.

        :return: A new :class:`.Listing` object that has been registered, if failure then return None.
        :type: :class:`.Listing` class

        For example::

            asset = DataAsset.create('test data asset', 'Some test data')
            listing_data = { 'price': 3.457, 'description': 'my data is for sale' }
            agent = SurferAgent(ocean)
            listing = agent.register_asset(asset,listing_data, account)

            if listing:
                print(f'registered my listing asset for sale with the did {listing.did}')

        """
        adapter = self._get_adapter()

        if self._did is None:
            raise ValueError('The agent must have a valid did')

        if not isinstance(listing_data, dict):
            raise ValueError('You must provide a dict as the listing data')

        listing = None

        register_data = adapter.register_asset(asset.metadata_text)
        if register_data:
            asset_id = register_data['asset_id']
            did = f'{self._did}/{asset_id}'
            asset.set_did(did)
            data = adapter.create_listing(asset_id, listing_data)
            listing = Listing(self, data['id'], asset, data)
        return listing

    def validate_asset(self, asset):
        """

        Validate an asset

        :param asset: Asset to validate.
        :return: True if the asset is valid
        """
        pass

    def upload_asset(self, asset):

        if not isinstance(asset, DataAsset):
            raise TypeError('Only DataAsset is supported')
        if not asset.data:
            raise ValueError('No data to upload')

        adapter = self._get_adapter()
        url = self.get_asset_store_url(asset.asset_id)
        return adapter.upload_asset_data(url, asset.asset_id, asset.data)


    def download_asset(self, asset_id, url):
        """
        Download an asset

        :param str url: url of the asset_id.

        :return: an Asset
        :type: :class:`.MemoryAsset` class

        """
        adapter = self._get_adapter()
        data = adapter.download_asset(url)
        store_asset = self.get_asset(asset_id)
        asset = DataAsset(
            store_asset.metadata,
            store_asset.did,
            data=data,
            metadata_text=store_asset.metadata_text
        )
        return asset

    def get_asset_store_url(self, asset_id):
        adapter = self._get_adapter()
        return adapter.get_asset_store_url(remove_0x_prefix(asset_id))

    def get_listing(self, listing_id):
        """
        Return an listing on the listings id.

        :param str listing_id: Id of the listing.

        :return: a listing object
        :type: :class:`.Asset` class

        """
        adapter = self._get_adapter()
        listing = None
        data = adapter.get_listing(listing_id)
        if data:
            asset_id = data['assetid']
            read_metadata = adapter.read_metadata(asset_id)
            if read_metadata:
                asset = self._create_asset_from_read(asset_id, read_metadata)
                listing = Listing(self, data['id'], asset, data)
        return listing

    def get_asset(self, asset_id):
        """

        This is for compatability for Surfer calls to get an asset directly from Surfer

        :param str asset_id: Id of the asset to read

        :return: an asset class
        :type: :class:`.AssetBase` class

        """
        asset = None
        adapter = self._get_adapter()
        clean_asset_id = remove_0x_prefix(asset_id)
        read_metadata = adapter.read_metadata(clean_asset_id)
        if read_metadata:
            asset = self._create_asset_from_read(asset_id, read_metadata)
        return asset

    def get_listings(self):
        """
        Returns all listings

        :return: List of listing objects

        """
        adapter = self._get_adapter()
        listings = []
        listings_data = adapter.get_listings()
        if listings_data:
            for data in listings_data:
                asset_id = data['assetid']
                read_metadata = adapter.read_metadata(asset_id)
                if read_metadata:
                    asset = self._create_asset_from_read(asset_id, read_metadata)
                    listing = Listing(self, data['id'], asset, data)
                    listings.append(listing)
        return listings

    def get_job(self, job_id):
        """

        Get a job from the invoke service ( koi )

        :param str job_id: Id of the job to get

        :return: a job class
        :type: :class:`starfish.job.Job` class

        """
        job = None
        adapter = self._get_adapter()
        data = adapter.get_job(job_id)
        if data:
            status = data.get('status', None)
            results = data.get('results', None)
            job = Job(job_id, status, results)
        return job

    def job_wait_for_completion(self, job_id, timeout_seconds=60, sleep_seconds=1):
        """

        Wait for a job to complete, with optional timeout seconds.

        :param int job_id: Job id to wait for completion
        :param int timeout_seconds: optional time in seconds to wait for job to complete, defaults to 60 seconds
        :param int sleep_seconds: optional number of seconds to sleep between polling the job status, == 0 no sleeping

        :return: Job object if finished, else False for timed out
        :type: :class:`starfish.job.Job` class or False

        """
        timeout_time = time.time() + timeout_seconds
        while timeout_time > time.time():
            job = self.get_job(job_id)
            if job.is_finished:
                return job
            if sleep_seconds > 0:
                time.sleep(sleep_seconds)
        return False

    def update_listing(self, listing):
        """

        Update the listing to the agent server.

        :param listing: Listing object to update
        :type listing: :class:`.Listing` class

        """
        adapter = self._get_adapter()
        return adapter.update_listing(listing.listing_id, listing.data)

    def search_listings(self, text, sort=None, offset=100, page=0):
        """

        Search the off chain storage for an asset with the givien 'text'

        :param str text: Test to search all metadata items for.
        :param sort: sort the results ( defaults: None, no sort).
        :type sort: str or None
        :param int offset: Return the result from with the maximum record count ( defaults: 100 ).
        :param int page: Returns the page number based on the offset.

        :return: a list of assets objects found using the search.
        :type: list of DID strings

        For example::

            # return the 300 -> 399 records in the search for the text 'weather' in the metadata.
            my_result = agent.search_registered_assets('weather', None, 100, 3)

        """
        # TODO: implement search listing in remote
        pass

    def purchase_asset(self, listing, account, purchase_id=None, status=None,
                       info=None, agreement=None):
        """

        Purchase an asset using it's listing and an account.

        :param listing: Listing to use for the purchase.
        :type listing: :class:`.Listing`
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param purchase_id: purchase id (optional)
        :type purchase_id: str or None
        :param status: purchase status (optional)
        :type status: str or None
        :param info: purchase info (optional)
        :type info: dict or None
        :param agreement: purchase agreement (optional)
        :type agreement: dict or None

        """
        adapter = self._get_adapter()
        purchase = {'listingid': listing.listing_id}
        if purchase_id:
            purchase['id'] = purchase_id
        if status:
            purchase['status'] = status
        if info:
            purchase['info'] = info
        if agreement:
            purchase['agreement'] = agreement
        return adapter.purchase_asset(purchase)

    def is_access_granted_for_asset(self, asset, account, purchase_id=None):
        """

        Check to see if the account and purchase_id have access to the assed data.


        :param asset: Asset to check for access.
        :type asset: :class:`.Asset` object
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str purchase_id: purchase id that was used to purchase the asset.

        :return: True if the asset can be accessed and consumed.
        :type: boolean
        """
        return False

    def get_asset_purchase_ids(self, asset):
        """

        Returns as list of purchase id's that have been used for this asset

        :param asset: Asset to return purchase details.
        :type asset: :class:`.Asset` object

        :return: list of purchase ids
        :type: list

        """
        return []

    def purchase_wait_for_completion(self, asset, account, purchase_id, timeoutSeconds):
        """

            Wait for completion of the purchase

            TODO: issues here...
            + No method as yet to pass back paramaters and values during the purchase process
            + We assume that the following templates below will always be used.

        """
        pass

    def consume_asset(self, listing, account, purchase_id):
        """
        Consume the asset and download the data. The actual payment to the asset
        provider will be made at this point.

        :param listing: Listing that was used to make the purchase.
        :type listing: :class:`.Listing`
        :param str purchase_id: purchase id that was used to purchase the asset.
        :param account: Ocean account that was used to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str download_path: path to store the asset data.

        :return: True if the asset has been consumed and downloaded
        :type: boolean

        """
        return False

    def invoke_result(self, asset, params=None, is_async=False):
        """

        Call an operation asset with params to execute a remote call.

        :param asset: Operation asset to use for this invoke call
        :type asset: :class:`.OperationAsset`
        :param dict params: Parameters to send to the invoke call

        :return: Return a dict of the result.
        :type: dict


        """

        if not isinstance(asset, OperationAsset):
            raise ValueError('Asset is not a OperationAsset')

        mode_type = 'async' if is_async else 'sync'
        if not asset.is_mode(mode_type):
            raise TypeError(f'This operation asset does not support {mode_type}')
        if not params:
            params = {}
        adapter = self._get_adapter()
        response = adapter.invoke(remove_0x_prefix(asset.asset_id), params, is_async)
        return response

    def get_endpoint(self, name):
        """

        Return the endpoint of the service available for this agent
        :param str name: name or type of the service, e.g. 'metadata', 'storage', 'Ocean.Meta.v1'
        """
        adapter = self._get_adapter()
        service_type = SurferAgentAdapter.find_supported_service_type(name)
        if service_type is None:
            raise ValueError(f'This agent does not support the following service name or type {name}')
        return adapter.get_endpoint(name)


    def _get_adapter(self, did=None, ddo=None, authorization=None):
        """

        Return a new SurferAgentAdapter object based on the did.
        If did == None then use the loaded did in this class.
        else check to see if the did != self._did, if not then load in the ddo as well
        """

        # if the given did is different, and no ddo, then we are requesting
        # data from a different source, so load in the ddo
        if did and did != self._did and ddo is None:
            if self._did and not self._ddo:
                squid_adapter = SquidAgentAdapter(self._ocean)
                ddo = squid_adapter.resolve_did_to_ddo(self._did)

        if did is None:
            did = self._did

        if ddo is None:
            ddo = self._ddo

        # TODO: check that the ddo is valid with the did
        if self._authorization is None:
            options = {
                'authorization': authorization
            }
        else:
            options = {
                'authorization': self._authorization
            }

        return SurferAgentAdapter(self._ocean, did, ddo, options)

    def _resolve_ddo_from_did(self, did):
        squid_adapter = SquidAgentAdapter(self._ocean)
        ddo_text = squid_adapter.resolve_did(self._did)
        if not ddo_text:
            raise ValueError(f'cannot find registered agent at {did}')
        return StarfishDDO(json_text=ddo_text)

    @property
    def did(self):
        """

        Return the did for this remote agent.

        :return: did of the registered agent
        :type: string
        """
        return self._did
    @property
    def ddo(self):
        """

        Return the registered DDO for this agent

        :return: DDO registered for this agent
        :type: :class:`.StarfishDDO`
        """
        return self._ddo

    @staticmethod
    def is_did_valid(did):
        """
        Checks to see if the DID string is a valid DID for this type of address for an asset.
        This method only checks the syntax of the DID, it does not resolve the DID
        to see if it is assigned to a valid Asset.

        :param str did: DID string to check to see if it is in a valid format.

        :return: True if the DID is in the format 'did:dep:xxxxx/yyyy'
        :type: boolean
        """
        data = did_parse(did)
        return data['path'] and data['id_hex']

    @staticmethod
    def decode_asset_did(did):
        data = did_parse(did)
        return data['id'], data['path']

    @staticmethod
    def generate_ddo(base_url_or_services, service_list=None):
        """
        Generate a DDO for the remote agent url. This DDO will contain the supported
        endpoints for the remote agent

        :param str base_url_or_services: Base URL of the remote agent
        :param dict base_url_or_services: Service dict to use. This has to be in the sample format:
            see starfish.agent.services SERVICES
        :param :class:`agent.services.Services` base_url_or_services: An agent Services object.

        :param dict service_list: Optional list of services to regsiter. This is only if the
            base_url_or_services is a string containing the base_url.

        :return: created DDO object assigned to the url of the remote agent service
        :type: :class:.`DDO`

            generate_ddo('http://localhost', ['metadata', 'storage'])

            # is the same as
            services = Services('http://localhost', service_list=['metadata', 'storage'])
            generate_ddo(services)

            # or

            generate_ddo(services.as_dict)

        """

        service_items = None

        if isinstance(base_url_or_services, str):
            services = Services(base_url_or_services, service_list)
            service_items = services.as_dict
        elif isinstance(base_url_or_services, dict):
            service_items = base_url_or_services
        elif isinstance(base_url_or_services, Services):
            service_items = base_url_or_services.as_dict
        else:
            raise(TypeError('Invalid services type, must be a string, dict or Services object'))

        did = did_generate_random()
        ddo = StarfishDDO(did)
        for _service_name, service_item in service_items.items():
            ddo.add_service(service_item['type'], service_item['url'], None)

        # add a signature
        private_key_pem = ddo.add_signature()
        # add the static proof
        ddo.add_proof(0, private_key_pem)

        return ddo

    @staticmethod
    def generate_metadata():
        return {"name": "string", "description": "string", "type": "dataset",
                "dateCreated": "2018-11-26T13:27:45.542Z",
                "tags": ["string"],
                "contentType": "string",
                "links": [{"name": "string", "type": "download", "url": "string"}]}

    def _create_asset_from_read(self, asset_id, read_metadata):
        # check the hash of the reading asset
        asset_id = remove_0x_prefix(asset_id)
        if not is_asset_hash_valid(asset_id, read_metadata['hash']):
            raise StarfishAssetInvalid(f' asset {asset_id} is not valid')
        did = f'{self._did}/{asset_id}'
        metadata_text = read_metadata['metadata_text']
        asset = create_asset_from_metadata_text(metadata_text, did)
        return asset



"""
    Starfish-py module

    This module contains the following classes

    * :class:`Account` The Account class.
    * :class:`Ocean` The main Ocean connection class.
    * :data:`logger` The logger class.
"""

__author__ = """DEX.sg"""
__version__ = '0.6.0'

import logging

logger = logging.getLogger(__name__)


from starfish.ocean.ocean import Ocean



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup
import os
from os.path import join

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.md') as changelog_file:
    changelog = changelog_file.read()

install_requirements = [
    'squid-py==0.7.1',
    'ocean-events-handler==0.1.2',
    'ocean-utils==0.1.0',
    'ocean-keeper==0.1.1',
    'ocean-secret-store-client==0.0.1',
    'requests==2.21.0',
    'plecos',
    'coloredlogs',
    'eciespy',
    'pyopenssl',
    'pycryptodome',
    'PyJWT',  # not jwt
    'PyYAML==4.2b4',
    'web3==4.5.0',
]

setup_requirements = ['pytest-runner', ]

test_requirements = [
    'codacy-coverage',
    'coverage',
    'docker',
    'flake8',
    'mccabe',
    'pyflakes',
    'pytest',
    'tox',
]

# Possibly required by developers of starfish-py:
dev_requirements = [
    'bumpversion',
    'pkginfo',
    'twine',
    'watchdog',
]

docs_requirements = [
    'Sphinx',
    'sphinx-rtd-theme',
    'sphinxcontrib-apidoc',
    'sphinxcontrib-plantuml',
    'sphinx-automodapi',
    'pygments',
]

packages = []
for d, _, _ in os.walk('starfish'):
    if os.path.exists(join(d, '__init__.py')):
        packages.append(d.replace(os.path.sep, '.'))

setup(
    author="dex-company",
    author_email='devops@dex.sg',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
    description="Ocean/squid-py wrapper.",
    extras_require={
        'test': test_requirements,
        'dev': dev_requirements + test_requirements + docs_requirements,
    },
    install_requires=install_requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='starfish-py',
    name='starfish-py',
    packages=packages,
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/DEX-Company/starfish-py',
    version='0.6.0',
    zip_safe=False,
)



from unittest.mock import Mock
import pytest
import secrets

from starfish import Ocean
from starfish.account import Account
from starfish.agent import RemoteAgent

def test_ocean_init(config):
    ocean = Ocean(
        network='nile',
        keeper_url=config.keeper_url,
        contracts_path=config.contracts_path,
        gas_limit=config.gas_limit,
        connect=False
    )
    assert(ocean)
    assert(ocean.keeper_url == config.keeper_url)
    assert(ocean.contracts_path == config.contracts_path)
    assert(ocean.gas_limit == config.gas_limit)

def test_ocean_init_empty(config):
    # now test with no block chain network
    ocean = Ocean()

    assert(ocean)

    account = ocean.load_account(config.accounts[0].address)

    # account should work even if there is no network
    assert( account)

    # error in register since account is in valid
    with pytest.raises(ValueError):
        info = ocean.register_did('did', 'ddo', account)

    assert(not ocean.search_operations('test search text') is None)

    assert(ocean.keeper_url == None)
    assert(ocean.contracts_path == None)
    assert(ocean.gas_limit == 0)


def test_register_update_agent_service(ocean, config):

    account = ocean.load_account(config.accounts[0].as_dict)

    ddo = RemoteAgent.generate_ddo(config.remote_agent_url)
    with pytest.raises(TypeError):
        ocean.register_did(ddo.did, ddo.as_text, None)

    with pytest.raises(TypeError):
        ocean.register_did(ddo.did, {}, account)

    reciept = ocean.register_did(ddo.did, ddo.as_text(), account)
    assert(reciept)

def test_search_operations(ocean):
    assert(not ocean.search_operations('test search text') is None)
    assert(len(ocean.search_operations('test search text')) == 0)
    assert(isinstance(ocean.search_operations('test search text'), list))

def test_load_account(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account)

"""
def test_accounts(ocean, config):
    accounts = ocean.accounts
    assert(accounts)
    assert(len(accounts) == len(config.accounts))
"""

def test_ocean_properties(ocean, config):
    assert(ocean.keeper_url == config.keeper_url)
    assert(ocean.contracts_path == config.contracts_path)
    assert(ocean.gas_limit == config.gas_limit)



from unittest.mock import Mock
import pytest
import secrets
import logging
import pathlib

from tests.integration.libs.integration_test_config import IntegrationTestConfig

from starfish import Ocean
from starfish.agent import (
    RemoteAgent,
    SquidAgent,
)

from starfish.agent.services import Services, ALL_SERVICES

INTEGRATION_PATH = pathlib.Path.cwd() / 'tests' / 'integration'
CONFIG_FILE_PATH = INTEGRATION_PATH / 'config.ini'

@pytest.fixture(scope="module")
def ocean():
    integrationTestConfig = IntegrationTestConfig(CONFIG_FILE_PATH)
    ocean = Ocean(
        keeper_url=integrationTestConfig.keeper_url,
        contracts_path=integrationTestConfig.contracts_path,
        gas_limit=integrationTestConfig.gas_limit,
        log_level=logging.WARNING
    )

    return ocean

@pytest.fixture(scope="module")
def config():
    integrationTestConfig = IntegrationTestConfig(CONFIG_FILE_PATH)
    return integrationTestConfig

@pytest.fixture(scope="module")
def remote_agent(ocean):
    integrationTestConfig = IntegrationTestConfig(CONFIG_FILE_PATH)

    ddo_options = None
    services = Services(integrationTestConfig.remote_agent_url, ALL_SERVICES)
    if integrationTestConfig.koi_url:
        services.add('invoke', f'{integrationTestConfig.koi_url}/api/v1')
    ddo = RemoteAgent.generate_ddo(services)
    options = {
        'url': integrationTestConfig.remote_agent_url,
        'username': integrationTestConfig.remote_agent_username,
        'password': integrationTestConfig.remote_agent_password,
    }
    return RemoteAgent(ocean, did=ddo.did, ddo=ddo, options=options)

@pytest.fixture(scope="module")
def squid_agent(ocean):
    integrationTestConfig = IntegrationTestConfig(CONFIG_FILE_PATH)
    return SquidAgent(ocean, integrationTestConfig.squid_config)

@pytest.fixture(scope='module')
def publisher_account(ocean, config):
    return ocean.load_account(config.publisher_account['address'], config.publisher_account['password'], config.publisher_account['keyfile'])


@pytest.fixture(scope='module')
def purchaser_account(ocean, config):
    return ocean.load_account(config.purchaser_account['address'], config.purchaser_account['password'], config.purchaser_account['keyfile'])


@pytest.fixture(scope='module')
def agent_account(ocean, config):
    return ocean.load_account(config.agent_account['address'], config.agent_account['password'], config.agent_account['keyfile'])



"""
    Memory Asset
"""
import os
from starfish.utils.crypto_hash import hash_sha3_256


from mimetypes import MimeTypes


from starfish.asset.asset_base import AssetBase

class DataAsset(AssetBase):
    """

    File asset can be used manage a data asset on the ocean network

    :param metadata: Dictionary metadata to provide for the asset.
    :type metadata: None or dict
    :param did: Optional did of the asset if it's registered
    :type did: None or str
    :param str data: Optional data of the asset, this can be str or bytes

    """
    def __init__(self, metadata, did=None, data=None,  metadata_text=None):
        if not isinstance(metadata, dict):
            raise ValueError('metadata must be a dict')

        if data:
            if not (isinstance(data, str) or isinstance(data, bytes)):
                raise ValueError('data can only be str or bytes')
            if not isinstance(data, bytes):
                data = data.encode('utf-8')

            metadata['contentHash'] = hash_sha3_256(data)
        metadata['type'] = 'dataset'

        self._data = data
        AssetBase.__init__(self, metadata, did, metadata_text)

    @staticmethod
    def create(name, data, metadata=None, did=None):
        """

        Create a new DataAsset using string or bytes data.

        :param str name: Name of the asset to create
        :param str, bytes data: Data to assign to the asset
        :param dict metadata: Optional metadata to add to the assets metadata
        :param str did: Option DID to assign to this asset

        :return: a new DataAsset
        :type: :class:`.DataAsset`

        """
        metadata = AssetBase.generateMetadata(name, 'dataset', metadata)
        if isinstance(data, str):
            metadata['contentType'] = 'text/plain; charset=utf-8'
        elif isinstance(data, bytes):
            metadata['contentType'] = 'application/octet-stream'
        else:
            raise ValueError('data can only be str or bytes')

        return DataAsset(metadata, did, data=data)

    @staticmethod
    def create_from_file(name, filename, metadata=None, did=None, is_read=True):
        """

        Create a new DataAsset using a file or filename.

        :param str name: Name of the asset to create
        :param str filename: If the filename is assigned to a valid file,
            the contents will be saved in the asset
        :param dict metadata: Optional metadata to add to the assets metadata
        :param str did: Option DID to assign to this asset
        :param bool is_read: If True read the file contents in as asset data.

        :return: a new DataAsset
        :type: :class:`.DataAsset`

        """

        metadata = AssetBase.generateMetadata(name, 'dataset', metadata)

        metadata['filename'] = str(filename)
        data = None
        if os.path.exists(filename):
            metadata['contentType'] = 'application/octet-stream'
            mime = MimeTypes()
            mime_type = mime.guess_type(f'file://{filename}')
            if mime_type:
                metadata['contentType'] = mime_type[0]
            if is_read:
                metadata['contentLength'] = os.path.getsize(filename)
                with open(filename, 'rb') as fp:
                    data = fp.read()
                metadata['size'] = len(data)

        return DataAsset(metadata, did, data=data)

    def save_to_file(self, filename):
        """
        Saves the data in the data asset to a file.

        :param str filename: Filename to save the data.

        """

        if self._data:
            with open(filename, 'wb') as fp:
                fp.write(self._data)

    @property
    def data(self):
        return self._data



"""


Agent Services


"""

from urllib.parse import urlparse, urljoin


SERVICES = {
    'metadata': {
        'type': 'DEP.Meta.{VERSION}',
        'url': '{BASE_URL}/api/{VERSION}/meta/data',
    },
    'storage': {
        'type': 'DEP.Storage.{VERSION}',
        'url': '{BASE_URL}/api/{VERSION}/assets',
    },
    'invoke': {
        'type': 'DEP.Invoke.{VERSION}',
        'url': '{BASE_URL}/api/{VERSION}/invoke',
    },
    'market': {
        'type': 'DEP.Market.{VERSION}',
        'url': '{BASE_URL}/api/{VERSION}/market',
    },
    'trust': {
        'type': 'DEP.Trust.{VERSION}',
        'url': '{BASE_URL}/api/{VERSION}/trust',
    },
    'auth': {
        'type': 'DEP.Auth.{VERSION}',
        'url': '{BASE_URL}/api/{VERSION}/auth',
    },
}

ALL_SERVICES=SERVICES.keys()

class Services:


    def __init__(self, base_url=None, service_list=None, version=None):
        self._base_url = base_url
        self._version = version
        if version is None:
            self._version = 'v1'
        self._items = {}
        if service_list:
            if self._base_url is None:
                raise ValueError('you must provide a base URL')
            for service_name in service_list:
                self.add(service_name)

    def add(self, name, url_or_uri=None, service_type=None, version=None):
        if url_or_uri is None and self._base_url is None:
            raise ValueError('you must provide a base URL or service URL')

        if name not in SERVICES:
            raise ValueError(f'service name {name} is not a valid service name')

        url = self._get_url(name, url_or_uri, version)
        service_type = self._get_service_type(name, service_type, version)
        self._items[name] = {
            'type': service_type,
            'url': url
        }

    def _get_url(self, name, url_or_uri=None, version=None):
        if version is None:
            version = self._version
        if url_or_uri is None:
            url_or_uri = SERVICES[name]['url'].format(BASE_URL=self._base_url, VERSION = version)
        url = url_or_uri
        info = urlparse(url)
        if not info.scheme:
            url = urljoin(self._base_url, url)
        return url

    def _get_service_type(self, name, service_type=None, version=None):
        if version is None:
            version = self._version
        if service_type is None:
            service_type = SERVICES[name]['type'].format(VERSION = version)

        return service_type

    @property
    def as_dict(self):
        return self._items

    @property
    def names(self):
        return self._items.keys()




from unittest.mock import Mock
import secrets

from tests.unit.libs.unit_test_account import UnitTestAccount

class UnitTestConfig():
    keeper_url = 'http://test-keeper-url:1234'
    contracts_path = 'test_artifacts_folder'
    gas_limit = 123456
    remote_agent_url = 'http://localhost:8080'
    accounts = {}
    def __init__(self):
        for index in range(0, 8):
            self.create_account(secrets.token_hex(32))

    def create_account(self, password):
        account = UnitTestAccount(password)
        self.accounts[len(self.accounts)] = account
        return account.address

unitTestConfig = UnitTestConfig()



import os
import logging
import json
import time

from unittest.mock import Mock
from eth_utils import add_0x_prefix

from squid_py import ConfigProvider
from squid_py.brizo.brizo import Brizo
from ocean_utils.did import id_to_did, did_to_id
from ocean_utils.agreements.service_agreement import ServiceAgreement
from ocean_utils.agreements.service_types import ServiceTypes
# from squid_py.agreements.manager import AgreementsManager

logging.getLogger('urllib3').setLevel(logging.WARNING)
logging.getLogger('web3').setLevel(logging.WARNING)

logger = logging.getLogger('test.mocks.brizo_mock')

class BrizoMock(object):

    def __init__(self, ocean_instance=None, account=None):
        self._ocean_instance = ocean_instance
        self._is_event_subscribed = False
        self._ddo_records = {}

    def initialize_service_agreement(self, did, agreement_id, service_definition_id,
                                     signature, account_address, purchase_endpoint):
        print(f'BrizoMock.initialize_service_agreement: purchase_endpoint={purchase_endpoint}')
        self._ocean_instance.agreements.create(
            did,
            service_definition_id,
            agreement_id,
            signature,
            account_address,
            self._account
        )
        return True

    @staticmethod
    def consume_service(service_agreement_id, service_endpoint, account_address, files,
                        destination_folder, *_, **__):
        for f in files:
            with open(os.path.join(destination_folder, os.path.basename(f['url'])), 'w') as of:
                of.write(f'mock data {service_agreement_id}.{service_endpoint}.{account_address}')

    @staticmethod
    def invoke_service(service_agreement_id, service_endpoint, account_address, invoke_payload):
        """
        Call the Koi endpoint to get access to the different files that form the asset.

        :param service_agreement_id: Service Agreement Id, str
        :param service_endpoint: Url to consume, str
        :param account_address: ethereum address of the consumer signing this agreement, hex-str
        :param files: List containing the files to be consumed, list
        :param destination_folder: Path, str
        :return:
        """
        logger.info(f'invoke endpoint with this url: {service_endpoint}')

        return invoke_payload['params']

    @staticmethod
    def get_brizo_url(config):
        return Brizo.get_brizo_url(config)


    @staticmethod
    def get_purchase_endpoint(config):
        return f'{Brizo.get_brizo_url(config)}/services/access/initialize'

    @staticmethod
    def get_service_endpoint(config):
        return f'{Brizo.get_brizo_url(config)}/services/consume'



"""
    test_07_metadata_access

    As a developer working with Ocean,
    I need a way to request metadata for an arbitrary asset

"""

import secrets
import json

from starfish.asset import DataAsset
from starfish.agent import RemoteAgent
from starfish.ddo.starfish_ddo import StarfishDDO



def test_07_metadata_access(ocean, resources, remote_agent):
    test_data = secrets.token_hex(1024)
    asset = DataAsset.create('TestAsset', test_data)
    listing = remote_agent.register_asset(asset, resources.listing_data)
    assert(not listing is None)
    assert(listing.listing_id)
    store_listing = remote_agent.get_listing(listing.listing_id)
    assert(store_listing)
    assert(store_listing.asset.asset_id)
    assert(store_listing.asset.metadata)
    assert(store_listing.asset.metadata == listing.asset.metadata)




import pytest
import secrets
from web3 import Web3

from starfish.account import account



def test_init(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account)
    assert(account.address)

    account = ocean.load_account(config.accounts[0].as_tuple)
    assert(account)
    assert(account.address)

    account = ocean.load_account(config.accounts[0].as_list)
    assert(account)
    assert(account.address)

def test_request_tokens(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    amount = secrets.randbelow(100) + 1
    assert(account.request_tokens(amount))



def test_is_address_equal(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account.is_address_equal(config.accounts[0].test_address))


def test_is_valid(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account.is_valid)

def test_is_password(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account.password == config.accounts[0].test_password)

def test_is_keyfile(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account.keyfile == config.accounts[0].test_keyfile)

def test_address(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account.address == config.accounts[0].test_address)

def test_as_checksum_address(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account.as_checksum_address == Web3.toChecksumAddress(config.accounts[0].test_address))

def test_password(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account.password == config.accounts[0].test_password)


def test_set_password(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    new_password = secrets.token_hex(48)
    account.set_password(new_password)
    assert(account.password != config.accounts[0].test_password)
    assert(account.password == new_password)


def test_ocean_balance(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account.ocean_balance == config.accounts[0].test_tokens)


def test_ether_balance(ocean, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    assert(account.is_valid)
    assert(account.ether_balance == config.accounts[0].test_ether)



import pytest


from starfish.keeper.direct_purchase import DirectPurchase

TOKEN_AMOUNT_TO_TRANSFER = 100
REFERENCE_1 = 4
REFERENCE_2 = 5

def test_direct_purchase(ocean, config, squid_agent, publisher_account, purchaser_account):
    purchaser_contract = DirectPurchase.get_instance()

    purchaser_account.request_tokens(TOKEN_AMOUNT_TO_TRANSFER)

    purchaser_balance = purchaser_account.ocean_balance
    publisher_balance = publisher_account.ocean_balance

    purchaser_contract.send_token_and_log(purchaser_account, publisher_account.address, TOKEN_AMOUNT_TO_TRANSFER, REFERENCE_1, REFERENCE_2)

    assert(purchaser_account.ocean_balance + TOKEN_AMOUNT_TO_TRANSFER == purchaser_balance
        and publisher_account.ocean_balance - TOKEN_AMOUNT_TO_TRANSFER == publisher_balance)

def test_is_paid(ocean, config, publisher_account, purchaser_account):
    purchaser_contract = DirectPurchase.get_instance()
    isPaid = purchaser_contract.check_is_paid(purchaser_account.address, publisher_account.address, TOKEN_AMOUNT_TO_TRANSFER, REFERENCE_2)
    assert(isPaid)



"""
    Test invoke agent

"""

import pathlib
import json
import logging
import time
from web3 import Web3

from starfish import (
    Ocean,
    logger
)
from starfish.agent import SquidAgent
from starfish.asset import DataAsset

from ocean_utils.agreements.service_factory import ServiceDescriptor
from ocean_utils.utils.utilities import generate_new_id
from squid_py.brizo.brizo_provider import BrizoProvider
from squid_py.brizo.brizo import Brizo
from tests.integration.mocks.brizo_mock import BrizoMock
import requests
import hashlib
from starfish import (
    Ocean,
    logger
)
from starfish.agent.invoke_agent import InvokeAgent

logger = logging.getLogger('test.invoke_oa')


def _register_asset_for_sale(agent, account):
    asset = DataAsset.create('TestAsset')
    listing = agent.register_asset(asset, account=account)
    assert listing
    assert listing.asset.did
    return listing

def purchase_asset(ocean, config, brizo_mock):


    agent = SquidAgent(ocean, config.squid_config)
    assert agent


    # test node has the account #0 unlocked
    publisher_account = ocean.get_account(config.publisher_account)
    publisher_account.unlock()
    publisher_account.request_tokens(20)

    listing = _register_asset_for_sale(agent, publisher_account)
    assert listing
    assert publisher_account

    listing_did = listing.asset.did
    # start to test getting the asset from storage
    listing = agent.get_listing(listing_did)
    assert listing
    assert listing.asset.did == listing_did
    ld=listing.data
    logger.info(f' listing.data is {ld}')
    purchase_account = ocean.get_account(config.purchaser_account)
    logger.info(f'purchase_account {purchase_account.ocean_balance}')

    purchase_account.unlock()

    purchase_account.request_tokens(10)

    time.sleep(2)
    logger.info(f'purchase_account after token request {purchase_account.ocean_balance}')

    brizo_mock.subscribe(ocean, publisher_account._squid_account)

    # test purchase an asset
    purchase_asset = listing.purchase(purchase_account)
    assert purchase_asset

    if not brizo_mock.is_event_subscribed:
        brizo_mock.subscribe(ocean, publisher_account._squid_account)

    assert(not purchase_asset.is_completed(purchase_account))

    error_message = purchase_asset.wait_for_completion(purchase_account)
    assert(error_message == True)

    assert(purchase_asset.is_completed(purchase_account))

    assert purchase_asset.is_purchased
    assert purchase_asset.is_purchase_valid(purchase_account)
    return purchase_asset.purchase_id,listing_did


def _test_invoke_with_sa(ocean, config):

    said,did=purchase_asset(ocean, config)
    assert said
    agent = InvokeAgent()
    assert agent

    res=agent.get_operations()
    assert 'hashing_did'==res['hashing']

    op=agent.get_operation('hashing_did')
    assert op

    sch=op.get_schema()
    assert 1==len(sch)
    assert sch['to_hash']=='asset'
    url='http://samplecsvs.s3.amazonaws.com/Sacramentorealestatetransactions.csv'
    download = requests.get(url)
    m = hashlib.sha256()
    m.update(download.content)
    hashval= m.hexdigest()

    res=op.invoke(to_hash={'serviceAgreementId':said,
        'did':did,
        'url':url,
        'consumerAddress':config.purchaser_account['address']})
    logger.info(f' invoke returns {res}')
    # TO DO: This needs testing again, since koi fails on calling invoke in the current barge.
    # assert res['hash']==hashval



"""
    test_09_construct_service_agreement

    As a developer working with Ocean,
    I need a way to create a valid service agreement that I can offer to others for puchase

"""




def test_09_construct_service_agreement():
    pass



#!/usr/bin/env python
"""

    Tests for `ocean CLI`.

"""
#import unittest
#import random
import os
import os.path
import sys
import subprocess
import re
# import logging

# from pytest import (
#     mark,
#     raises,
# )

def test_ocean_cli():
    log_msg = 'Using default logging settings.'
    header = 'Index     Account                                             Ocean Tokens                                   Ether'
    program = sys.argv[0]
    if program[0] != '/':
        program = os.path.join(os.getcwd(), program)

    # fix for running from the command line 'python -m pytest tests' within venv
    if re.search(r'/venv/', program):
        program = './'

    pdir = os.path.normpath(os.path.dirname(program))
    dot_tox = pdir.find('/.tox')
    if dot_tox > 0:
        pdir = pdir[0:dot_tox]
    cli_path = os.path.join(pdir, "ocean")
    command = "balance"
    args = [cli_path, '--config', 'tests/resources/config.ini', command]
    cli = subprocess.run(args, stdout=subprocess.PIPE)
    stdout = cli.stdout.decode()
    assert 0 == cli.returncode
    lines = stdout.split('\n')
    print(lines)
    num_accounts = 3
    assert (num_accounts + 3) == len(lines)
    if len(lines) == (num_accounts + 3):
        assert log_msg == lines[0]
        assert header == lines[1]
        for i in range(num_accounts):
            account = lines[i + 2].split()
            assert 4 == len(account)
        last_account_id = account[1]
        # FUTURE assert valid_id(last_account_id)
        assert 0 == len(lines[num_accounts + 2])
    # test with narrowing to just one account
    args.append(last_account_id)
    cli = subprocess.run(args, stdout=subprocess.PIPE)
    stdout = cli.stdout.decode()
    assert 0 == cli.returncode
    lines = stdout.split('\n')
    num_accounts = 1
    assert (num_accounts + 3) == len(lines)
    if len(lines) == (num_accounts + 3):
        assert log_msg == lines[0]
        assert header == lines[1]
        for i in range(num_accounts):
            account = lines[i + 2].split()
            assert 4 == len(account)
        last_account_id2 = account[1]
        # FUTURE assert valid_id(last_account_id2)
        assert 0 == len(lines[num_accounts + 2])
    assert last_account_id == last_account_id2



from unittest.mock import Mock
import pytest
import secrets
import json
import re
from web3 import Web3

from starfish.ddo.starfish_ddo import StarfishDDO
from starfish.middleware.squid_agent_adapter import SquidAgentAdapterPurchaseError

from ocean_utils.did import (
    did_to_id,
    id_to_did,
    did_to_id_bytes,
)
from ocean_utils.agreements.service_types import ServiceTypes
from squid_py.ocean.keeper import SquidKeeper
from plecos import is_valid_dict_local, validate_dict_local


from tests.unit.libs.unit_test_config import unitTestConfig

TEST_SERVICE_NAME = 'service_data'

class MockKeeper():
    def sign_hash(self, text, publisher_account):
        return Web3.sha3(text=text + publisher_account.address)


class MockSquidAgentAdapter():
    def __init__(self, ocean, options=None):
        """init a standard ocean object"""
        self._ocean = ocean
        self._options = options
        self._metadata = {}
        self._ddo_list = {}
        self._purchase_assets={}
        self._account_list = {}

    def get_account(self, address, password=None, keyfile=None):
        if address:
            for index in unitTestConfig.accounts:
                test_account = unitTestConfig.accounts[index]
                if address.lower() == test_account.test_address.lower():
                    account = Mock()
                    account.address = address
                    if password:
                        account.password = password
                    if keyfile:
                        account.keyfile = keyfile
                    self._account_list[address] = account
                    return account
        return None

    def register_agent(self, service_name, endpoint_url, account, did=None):
        # if no did then we need to create a new one
        did = id_to_did(secrets.token_hex(32))

        # create a new DDO
        ddo = StarfishDDO(did)
        # add a signature
        private_key_pem = ddo.add_signature()
        # add the service endpoint with the meta data
        ddo.add_service(service_name, endpoint_url)
        # add the static proof
        ddo.add_proof(0, private_key_pem)
        # if self.register_ddo(did, ddo, account._squid_account):
        return [did, ddo, private_key_pem]

    def validate_metadata(self, metadata):
        """

        Validate the metadata with plesto

        :param dict metadata: metadata to validate
        :return: True if the metadata is valid
        :type: boolean

        """
        if is_valid_dict_local(metadata):
            return True
        else:
            validator = validate_dict_local(metadata)

        return False

    def register_ddo(self, did, ddo, account):
        self._ddo_list[did] = ddo
        return secrets.token_hex(32)

    def resolve_did(self, did):
        return self._ddo_list[did]

    def register_asset(self, metadata, account ):
        did = id_to_did(secrets.token_hex(32))
        self._metadata[did] = metadata

        # create a new DDO
        ddo = StarfishDDO(did)
        # add a signature
        private_key_pem = ddo.add_signature()
        # add the service endpoint with the meta data
        assert(self._options)
        assert('aquarius_url' in self._options)
        ddo.add_service(TEST_SERVICE_NAME, self._options['aquarius_url'])
        ddo.add_service(ServiceTypes.METADATA, '', {'metadata': metadata})
        # add the static proof

        mockKeeper = MockKeeper()

        if not 'checksum' in metadata['base']:
            metadata['base']['checksum'] = did

        ddo.add_proof_keeper(metadata['base']['checksum'], account, mockKeeper)
        # if self.register_ddo(did, ddo, account._squid_account):
        self._ddo_list[did] = ddo
        return ddo

    def search_assets(self, text, sort=None, offset=100, page=0):
        result = []
        for ddo in self._ddo_list.values():
            metadata_text = json.dumps(ddo.metadata)
            if re.search(text, metadata_text):
                result.append(ddo)

        return result

    def read_asset(self, did):
        return self._ddo_list[did]

    def purchase_asset(self, ddo, account):
        service = ddo.get_service(TEST_SERVICE_NAME)
        assert(service)
        service_dict = service.as_dictionary()
        assert(service_dict['serviceEndpoint'] == self.options['aquarius_url'])
        service_agreement_id = secrets.token_hex(32)
        self._purchase_assets[service_agreement_id] = ddo.did
        return service_agreement_id

    def purchase_wait_for_completion(self, asset, account, purchase_id, timeoutSeconds):
        if purchase_id:
            return True
        raise SquidAgentAdapterPurchaseError('test squid agent adapter purchase wait error')

    def consume_asset(self, ddo, account, service_agreement_id):
        service = ddo.get_service(TEST_SERVICE_NAME)
        assert(service)
        service_dict = service.as_dictionary()
        assert(service_dict['serviceEndpoint'] == self.options['aquarius_url'])
        assert(service_agreement_id in self._purchase_assets)
        did = self._purchase_assets[service_agreement_id]
        if not did in self._metadata:
            return False
        return self._metadata[did]['base']['files']


    def is_access_granted_for_asset(self, did, account, service_agreement_id=None):
        if did in self._metadata and service_agreement_id in self._purchase_assets:
            return True
        return False

    def get_account_balance(self, account):
        for index in unitTestConfig.accounts:
            test_account = unitTestConfig.accounts[index]
            if account.address.lower() == test_account.test_address.lower():
                balance = Mock()
                balance.eth = Web3.toWei(test_account.test_ether, 'ether')
                balance.ocn = Web3.toWei(test_account.test_tokens, 'ether')
                return balance
        return 0

    @property
    def accounts(self):
        """
        Not used at the moment
        """
        return self._account_list

    def request_tokens(self, value, account):
        found_account = self.get_account(account.address, account.password, account.keyfile)
        if found_account.address == account.address:
            return value
        return 0

    """
    @property
    def accounts(self):
        result = []
        for index in unitTestConfig.accounts:
            test_account = unitTestConfig.accounts[index]
            account = Mock()
            account.address = test_account.test_address,
            account.password = test_account.test_password
            result.append(account)

        return result
    """

    @property
    def options(self):
        return self._options



"""
    test_pd_case_search_listings

    As a developer
    I want register an asset in squid, and then find it in the search listings,

"""

import secrets
import logging
import json
from web3 import Web3

from starfish.asset import (
    DataAsset,
    RemoteDataAsset,
)

def test_pd_case_file_transfer(ocean, config, resources, remote_agent, squid_agent, publisher_account):

    # for the use case , we need to assign a unique id to the asset
    # so we know that it's from the publisher
    pd_test_case_tag = 'pd_test_case'
    unique_pd_case_id = secrets.token_hex(32)
    dummy_did = secrets.token_hex(32)
    dummy_asset_id = secrets.token_hex(32)
    # create a pretend dummy url based on the dummy remote agent DID & dummy asset_id saved in remote agent
    dummy_url = 'op:did:{dummy_did}/{dummy_asset_id}'

    # now create a checksum so that we know this actually comes from the correct publisher
    valid_check = Web3.toHex(Web3.sha3(text=f'{unique_pd_case_id}{dummy_url}{publisher_account.address}'))

    listing_data = {
        'name': 'Test file asset',
        'description': 'Test asset for sale - not for public purchase',
        'author': 'pd_test_case asset listing',
        'license': 'Closed',
        'price': 3.141592,
        'extra_data': {
            'id': unique_pd_case_id,
            'valid_check': valid_check,
        },
        'tags': [pd_test_case_tag],
    }
    asset_sale = RemoteDataAsset.create_with_url('TestAsset', dummy_url)
    # print('metadata ',squid_agent._convert_listing_asset_to_metadata(asset_sale, listing_data))

    listing = squid_agent.register_asset(asset_sale, listing_data, account=publisher_account)
    assert(listing)

    listing_items = squid_agent.search_listings({'tags': [pd_test_case_tag]})

    assert(len(listing_items) >= 1)
    for listing in listing_items:
        assert('tags' in listing.data and pd_test_case_tag in listing.data['tags'])
        assert(listing.data['extra_data'])
        assert(isinstance(listing.data['extra_data'], dict))
        if listing.data['extra_data']['id'] == unique_pd_case_id:
            assert(listing.data['extra_data']['valid_check'] == valid_check)
